"""A group of score reels."""
from collections import deque

import asyncio

from mpf.core.system_wide_device import SystemWideDevice
from mpf.devices.score_reel_controller import ScoreReelController


class ScoreReelGroup(SystemWideDevice):

    """Represents a logical grouping of score reels in a pinball machine.

    Multiple individual ScoreReel object make up the individual digits of this
    group. This group also has support for the blank zero "inserts" that some
    machines use. This is a subclass of mpf.core.device.Device.
    """

    config_section = 'score_reel_groups'
    collection = 'score_reel_groups'
    class_label = 'score_reel_group'

    @classmethod
    def device_class_init(cls, machine):
        """If we have at least one score reel group, we need a ScoreReelController."""
        machine.score_reel_controller = ScoreReelController(machine)

    def __init__(self, machine, name):
        """Initialise score reel group."""
        super().__init__(machine, name)

        self.wait_for_valid_queue = None
        self.valid = True  # Confirmed reels are showing the right values
        self.unlight_on_resync_key = None
        self.light_on_valid_key = None

        self.reels = []
        # A list of individual ScoreReel objects that make up this
        # ScoreReelGroup. The number of items in the list correspondis to the
        # number of digits that can be displayed. A value of `None` indicates a
        # position that is not controlled by a moving reel (like a fake ones
        # digit).

        # Note that this is "backwards," with element 0 representing the ones
        # digit, element 1 representing the tens, etc..

        self.desired_value_list = []
        # A list of what values the machine desires to have the score reel
        # group set to.

        self.advance_queue = deque()
        # Holds a list of the next reels that for step advances.

        self.jump_in_progress = False
        # Boolean attribute that is True when a jump advance is in progress.

        self._tick_task = None

    async def _initialize(self):
        await super()._initialize()
        self.reels = self.config['reels']
        self.reels.reverse()  # We want our smallest digit in the 0th element

        self.config['chimes'].reverse()
        for i in range(len(self.config['chimes'])):

            if self.config['chimes'][i]:
                if not self.reels[i]:
                    self.raise_config_error("Invalid reel for chime {}".format(self.config['chimes'][i]), 1)
                self.machine.events.add_handler(event='reel_' + self.reels[i].name + '_advance',
                                                handler=self.chime,
                                                chime=self.config['chimes'][i])

    @classmethod
    def chime(cls, chime, **kwargs):
        """Pulse chime."""
        del kwargs
        chime.pulse()

    def set_value(self, value):
        """Reset the score reel group to display the value passed.

        This method will "jump" the score reel group to display the value
        that's passed as an it. (Note this "jump" technique means it will just
        move the reels as fast as it can, and nonsensical values might show up
        on the reel while the movement is in progress.)

        This method is used to "reset" a reel group to all zeros at the
        beginning of a game, and can also be used to reset a reel group that is
        confused or to switch a reel to the new player's score if multiple
        players a sharing the same reel group.

        Note you can choose to pass either an integer representation of the
        value, or a value list.

        Args:
        ----
            value: An integer value of what the new displayed value (i.e. score)
                should be. This is the default option if you only pass a single
                positional argument, e.g. `set_value(2100)`.
        """
        value_list = self.int_to_reel_list(value)

        self.log.debug("Jumping to %s.", value_list)

        # set the new desired value which we'll use to verify the reels land
        # where we want them to.
        self.desired_value_list = value_list

        # loop through the reels one by one
        for i in range(len(self.reels)):
            if not self.reels[i]:
                continue

            self.reels[i].set_destination_value(self.desired_value_list[i])

    def wait_for_ready(self):
        """Return a future which will be done when all reels reached their destination."""
        futures = []
        for reel in self.reels:
            if reel:
                futures.append(reel.wait_for_ready())

        return asyncio.wait(iter(futures))

    def int_to_reel_list(self, value):
        """Convert an integer to a list of integers that represent each positional digit in this ScoreReelGroup.

        The list returned is in reverse order. (See the example below.)

        The list returned is customized for this ScoreReelGroup both in terms
        of number of elements and values of `None` used to represent blank
        plastic zero inserts that are not controlled by a score reel unit.

        For example, if you have a 5-digit score reel group that has 4
        phyiscial reels in the tens through ten-thousands position and a fake
        plastic "0" insert for the ones position, if you pass this method a
        value of `12300`, it will return `[None, 0, 3, 2, 1]`

        This method will pad shorter ints with zeros, and it will chop off
        leading digits for ints that are too long. (For example, if you pass a
        value of 10000 to a ScoreReelGroup which only has 4 digits, the
        returns list would correspond to 0000, since your score reel unit has
        rolled over.)

        Args:
        ----
            value: The interger value you'd like to convert.

        Returns a list containing the values for each corresponding score reel,
        with the lowest reel digit position in list position 0.
        """
        output_list = []

        # convert our number to a string
        str_value = str(value)

        # pad the string with leading zeros
        str_value = str_value.zfill(len(self.reels))

        # slice off excess characters if the value is longer than num of reels

        # how many digits do we have to slice?
        trim = len(str_value) - len(self.reels)
        # and... slice!
        str_value = str_value[trim:]

        # todo if we don't do the above trim then it will just show the highest
        # digits, effective "shifting" the score by one. Might be a fun feature?

        # generate our list with one digit per item
        for digit in str_value:
            output_list.append(int(digit))

        # reverse the list so the least significant is first
        output_list.reverse()

        # replace fake position digits with `None`
        for i, dummy_output in enumerate(output_list):
            if not self.reels[i]:
                output_list[i] = None

        return output_list

    def light(self, **kwargs):
        """Light up this ScoreReelGroup based on the 'light_tag' in its config."""
        del kwargs
        self.log.debug("Turning on Lights")
        for light in self.machine.lights.items_tagged(
                self.config['lights_tag']):
            light.on()

    def unlight(self, **kwargs):
        """Turn off the lights for this ScoreReelGroup based on the 'light_tag' in its config."""
        del kwargs
        self.log.debug("Turning off Lights")
        for light in self.machine.lights.items_tagged(
                self.config['lights_tag']):
            light.off()



"""Test gottlieb trough setup with system11 trough + drain device."""
from mpf.tests.MpfTestCase import MpfTestCase, test_config
from unittest.mock import MagicMock


class TestGottliebTrough(MpfTestCase):

    """Test Gottlieb style troughs with outhole and only one one or two switches."""

    def get_config_file(self):
        return 'test_gottlieb_trough.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def test_boot_with_ball_in_drain_empty_trough(self):
        # MPF starts with a ball in the outhole (drain device). It should be
        # ejected into the trough and stay there.
        self.machine.coils["outhole"].pulse = MagicMock()
        self.machine.coils["trough"].pulse = MagicMock()

        self.hit_switch_and_run("outhole", 0.6)

        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)

        self.release_switch_and_run("outhole", 1)

        self.hit_and_release_switch("trough_entry")
        self.advance_time_and_run(1)

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(1, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

    @test_config("test_gottlieb_trough_with_initial_balls.yaml")
    def test_boot_with_balls_in_drain_and_trough(self):
        self.machine.coils["outhole"].pulse = MagicMock()
        self.machine.coils["trough"].pulse = MagicMock()

        self.hit_switch_and_run("outhole", 1)
        self.advance_time_and_run(.6)

        # trough is full. there should be no eject
        self.assertEqual(0, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)

        self.assertEqual(1, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertNumBallsKnown(4)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual(0, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)
        # self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

        # start game
        self.hit_and_release_switch("start")
        self.advance_time_and_run(1)
        self.assertIsNotNone(self.machine.game)
        self.assertEqual(0, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(1, self.machine.coils["trough"].pulse.call_count)

        # now the trough has space and the outhole can eject
        self.release_switch_and_run("trough_entry", 7)
        self.assertEqual(0, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(1, self.machine.coils["trough"].pulse.call_count)

        self.hit_switch_and_run("plunger", 1)
        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(1, self.machine.coils["trough"].pulse.call_count)

        self.release_switch_and_run("outhole", 1)
        self.release_switch_and_run("plunger", 1)
        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(1, self.machine.coils["trough"].pulse.call_count)

        self.hit_switch_and_run("trough_entry", 4)

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)

        # ball drains
        self.hit_switch_and_run("outhole", 1)
        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(2, self.machine.coils["trough"].pulse.call_count)

        self.machine.log.warning("TEST: DRAIN")

        # we usually see the ball in the plunger first
        self.hit_switch_and_run("plunger", 1)
        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(2, self.machine.coils["trough"].pulse.call_count)

        # about a second later the trough switch deactives
        self.release_switch_and_run("trough_entry", 1)
        self.assertEqual(2, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(2, self.machine.coils["trough"].pulse.call_count)

        self.release_switch_and_run("outhole", 1)
        self.hit_switch_and_run("trough_entry", 1)

        self.release_switch_and_run("plunger", 10)
        self.assertEqual(2, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(2, self.machine.coils["trough"].pulse.call_count)

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)

    @test_config("test_gottlieb_trough_with_initial_balls.yaml")
    def test_add_ball_to_pf(self):
        self.machine.coils["outhole"].pulse = MagicMock()
        self.machine.coils["trough"].pulse = MagicMock()

        self.assertEqual(0, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)
        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        self.assertEqual(1, self.machine.coils["trough"].pulse.call_count)
        self.advance_time_and_run(.1)
        # simulate some bouncing on the switch on eject
        self.release_switch_and_run("trough_entry", 1)
        self.hit_switch_and_run("trough_entry", .3)
        self.release_switch_and_run("trough_entry", .6)

        self.machine.switch_controller.process_switch("plunger", 1)

        self.advance_time_and_run(1)

        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('ejecting', self.machine.ball_devices["plunger"]._state)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual(1, self.machine.ball_devices["playfield"].num_balls_requested)
        self.machine.switch_controller.process_switch("plunger", 0)

        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("playfield", 1)
        self.machine.switch_controller.process_switch("playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

        self.assertEqual(0, self.machine.ball_devices["playfield"].num_balls_requested)
        self.assertEqual(1, self.machine.ball_devices["playfield"].balls)

    @test_config("test_gottlieb_trough_with_initial_balls.yaml")
    def test_boot_and_start_game_with_ball_in_plunger(self):
        self.machine.coils["outhole"].pulse = MagicMock()
        self.machine.coils["trough"].pulse = MagicMock()

        self.machine.switch_controller.process_switch("plunger", 1)
        self.advance_time_and_run(.6)

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(1, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('ejecting', self.machine.ball_devices["plunger"]._state)

        # should not start
        self.hit_and_release_switch("start")
        self.advance_time_and_run(1)
        self.assertIsNone(self.machine.game)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)

        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("plunger", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)
        self.machine.switch_controller.process_switch("playfield", 1)
        self.machine.switch_controller.process_switch("playfield", 0)
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("playfield", 1)
        self.machine.switch_controller.process_switch("playfield", 0)
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("playfield", 1)
        self.machine.switch_controller.process_switch("playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(1, self.machine.ball_devices["playfield"].balls)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

        self.hit_switch_and_run("outhole", 1)

        # self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

        # game should start now
        self.hit_and_release_switch("start")
        self.advance_time_and_run(1)
        self.assertIsNotNone(self.machine.game)
        self.assertEqual(1, self.machine.coils["trough"].pulse.call_count)
        self.release_switch_and_run("trough_entry", 1)

        self.advance_time_and_run(.1)
        self.hit_switch_and_run("plunger", 1)
        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.release_switch_and_run("outhole", 1)
        self.hit_switch_and_run("trough_entry", 4)

        self.assertEqual(4, self.machine.ball_controller.num_balls_known)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('ejecting', self.machine.ball_devices["plunger"]._state)

        self.release_switch_and_run("plunger", 10)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

    def test_boot_with_two_balls_in_trough(self):
        # two balls are in trough
        self.machine.coils["outhole"].pulse = MagicMock()
        self.machine.coils["trough"].pulse = MagicMock()

        # but mpf does not know and assumes 0 in trough
        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(0, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual(0, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

        # starting a game should fail
        self.hit_and_release_switch("start")
        self.advance_time_and_run(1)
        self.assertEqual(0, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)
        self.assertIsNone(self.machine.game)

        # a ball is added in while machine is running
        self.hit_switch_and_run("outhole", 1)

        self.assertEqual(1, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(0, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)
        self.assertEqual('ejecting', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

        # outhole ejects and ball enters trough
        self.release_switch_and_run("outhole", 1)
        # ball three sits on entrance switch
        self.hit_switch_and_run("trough_entry", 4)

        # trough recognizes that its actually full
        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)
        self.assertEqual(1, self.machine.coils["outhole"].pulse.call_count)
        self.assertEqual(0, self.machine.coils["trough"].pulse.call_count)
        self.assertEqual('idle', self.machine.ball_devices["outhole"]._state)
        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)
        self.assertEqual(3, self.machine.ball_controller.num_balls_known)

        # game should now start
        self.hit_and_release_switch("start")
        self.advance_time_and_run(1)
        self.assertIsNotNone(self.machine.game)

    def test_single_ball_drain_and_eject(self):
        # tests that when a ball drains into the outhole and is in the process
        # of being ejected to the trough, MPF is able to also request a ball
        # to the plunger even though the ball hasn't made it into the trough
        # yet

        self.machine.ball_devices["trough"].ball_count_handler.counter._last_count = 3
        self.machine.ball_devices["trough"].available_balls = 3
        self.machine.ball_controller.num_balls_known = 3
        self.machine.ball_devices["trough"].ball_count_handler._set_ball_count(3)
        self.advance_time_and_run(1)

        self.machine.coils["trough"].pulse = MagicMock()

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(0, self.machine.ball_devices["playfield"].balls)

        # start a game, ball goes to plunger
        self.hit_and_release_switch("start")
        self.advance_time_and_run(5)
        self.assertIsNotNone(self.machine.game)

        self.machine.playfield.add_ball()
        self.machine.playfield.add_ball()

        self.assertEqual(1, self.machine.coils["trough"].pulse.call_count)

        self.machine.switch_controller.process_switch("plunger", 1)
        self.advance_time_and_run(2)

        # plunge to playfield
        self.machine.switch_controller.process_switch("plunger", 0)
        self.advance_time_and_run(2)
        self.machine.switch_controller.process_switch("playfield", 1)
        self.machine.switch_controller.process_switch("playfield", 0)
        self.advance_time_and_run(2)

        self.assertEqual(2, self.machine.coils["trough"].pulse.call_count)

        self.machine.switch_controller.process_switch("plunger", 1)
        self.advance_time_and_run(2)

        # plunge to playfield
        self.machine.switch_controller.process_switch("plunger", 0)
        self.advance_time_and_run(2)
        self.machine.switch_controller.process_switch("playfield", 1)
        self.machine.switch_controller.process_switch("playfield", 0)
        self.advance_time_and_run(3)

        self.assertEqual(3, self.machine.coils["trough"].pulse.call_count)

        self.machine.switch_controller.process_switch("plunger", 1)
        self.advance_time_and_run(2)

        # plunge to playfield
        self.machine.switch_controller.process_switch("plunger", 0)
        self.advance_time_and_run(2)
        self.machine.switch_controller.process_switch("playfield", 1)
        self.machine.switch_controller.process_switch("playfield", 0)
        self.advance_time_and_run(3)

        self.assertEqual(0, self.machine.ball_devices["outhole"].balls)
        self.assertEqual(0, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(3, self.machine.ball_devices["playfield"].balls)

        # drain
        self.machine.switch_controller.process_switch("outhole", 1)
        self.advance_time_and_run(.6)
        self.assertEqual(2, self.machine.ball_devices["playfield"].balls)
        self.assertEqual(2, self.machine.game.player.ball)

        self.machine.switch_controller.process_switch("outhole", 0)
        self.advance_time_and_run(.1)

    @test_config("test_gottlieb_trough_with_initial_balls.yaml")
    def test_drain_during_game_start(self):
        # A ball drains shortly after game start
        self.machine.coils["outhole"].pulse = MagicMock()
        self.machine.coils["trough"].pulse = MagicMock()

        self.hit_switch_and_run("start", 0.15)
        self.release_switch_and_run("start", 0.73)
        self.hit_switch_and_run("playfield", 0.123)
        self.release_switch_and_run("playfield", 0.325)
        self.hit_switch_and_run("outhole", 0.448)
        self.hit_switch_and_run("plunger", 0.4)     # this was originally 0.888s. will cause a new ball when > 0.5s
        self.release_switch_and_run("trough_entry", 3.244)
        self.release_switch_and_run("plunger", 0.322)
        self.hit_and_release_switch("playfield")
        self.advance_time_and_run(20)

        self.assertEqual(4, self.machine.ball_controller.num_balls_known)

        self.assertEqual('idle', self.machine.ball_devices["trough"]._state)
        self.assertEqual('idle', self.machine.ball_devices["plunger"]._state)

    @test_config("test_gottlieb_trough_with_initial_balls.yaml")
    def test_double_drain_during_trough_eject(self):
        self.mock_event("balldevice_outhole_ball_eject_failed")
        self.mock_event("balldevice_ball_missing")
        self.machine.ball_devices["plunger"].config['eject_timeouts'][self.machine.playfield] = 20000
        self.machine.coils["outhole"].pulse = MagicMock()
        self.machine.coils["trough"].pulse = MagicMock()
        self.assertEqual(3, self.machine.ball_controller.num_balls_known)

        self.machine.ball_controller.num_balls_known = 7
        self.machine.playfield.available_balls = 3
        self.machine.playfield.balls = 3

        self.assertEqual(7, self.machine.ball_controller.num_balls_known)

        self.machine.playfield.add_ball(1)
        self.advance_time_and_run(1)

        self.release_switch_and_run("trough_entry", .5)
        self.hit_switch_and_run("plunger", 1)
        self.release_switch_and_run("plunger", 3)

        # make sure the eject does not fail when the trough has scheduled an eject
        self.machine.playfield.add_ball(1)

        self.hit_switch_and_run("outhole", 1)
        self.release_switch_and_run("outhole", 1)
        # a new ball enters outhole from the playfield
        self.hit_switch_and_run("outhole", 0)
        self.hit_switch_and_run("trough_entry", 4)
        self.advance_time_and_run(4)
        self.assertEventNotCalled("balldevice_outhole_ball_eject_failed")
        self.assertEventNotCalled("balldevice_ball_missing")

    @test_config("test_gottlieb_trough_with_initial_balls.yaml")
    def test_eject_during_incoming_ball(self):
        self.mock_event("balldevice_outhole_ball_eject_failed")
        self.mock_event("balldevice_ball_missing")
        self.machine.ball_devices["plunger"].config['eject_timeouts'][self.machine.playfield] = 20000
        self.machine.coils["outhole"].pulse = MagicMock()
        self.machine.coils["trough"].pulse = MagicMock()
        self.assertEqual(3, self.machine.ball_controller.num_balls_known)

        self.machine.ball_controller.num_balls_known = 7
        self.machine.playfield.available_balls = 3
        self.machine.playfield.balls = 3

        self.assertEqual(7, self.machine.ball_controller.num_balls_known)

        self.machine.playfield.add_ball(1)
        self.advance_time_and_run(1)

        self.release_switch_and_run("trough_entry", .5)
        self.hit_switch_and_run("plunger", 1)
        self.release_switch_and_run("plunger", 30)

        # a ball drains
        self.hit_switch_and_run("outhole", 1)
        self.release_switch_and_run("outhole", 1)

        # another ball enters the outhole
        self.hit_switch_and_run("outhole", 0)
        # and the trough gets an eject request
        self.machine.playfield.add_ball(1)
        self.advance_time_and_run()
        self.hit_and_release_switch("trough_entry")
        # this goed beyond the timeout of the outhole
        self.advance_time_and_run(6)
        # a new ball enters outhole from the playfield
        self.hit_switch_and_run("plunger", 1)
        self.release_switch_and_run("plunger", 1)

        self.assertEventNotCalled("balldevice_outhole_ball_eject_failed")
        self.assertEventNotCalled("balldevice_ball_missing")



import asyncio

from mpf.platforms.rpi import rpi
from mpf.tests.MpfTestCase import MpfTestCase


class MockApigpio():

    # gpio Pull Up Down

    PUD_OFF = 0
    PUD_DOWN = 1
    PUD_UP = 2

    # gpio modes

    INPUT = 0
    OUTPUT = 1

    # gpio edges

    RISING_EDGE = 0
    FALLING_EDGE = 1
    EITHER_EDGE = 2

    class Pi():

        def __init__(self, loop=None):
            del loop
            self.callbacks = {}
            self.modes = {}
            self.pull_ups = {}
            self.servos = {}
            self.outputs = {}
            self.i2c_write = []
            self.i2c_read = []

        async def connect(self, address):
            pass

        async def read_bank_1(self):
            ## only switch 1 is active
            return 0x02

        async def set_pull_up_down(self, gpio, pud):
            self.pull_ups[gpio] = pud

        async def set_mode(self, gpio, mode):
            self.modes[gpio] = mode

        async def add_callback(self, user_gpio, edge=0, func=None):
            self.callbacks[user_gpio] = func

        async def set_servo_pulsewidth(self, user_gpio, pulsewidth):
            self.servos[user_gpio] = pulsewidth

        async def write(self, gpio, level):
            self.outputs[gpio] = level

        async def set_PWM_dutycycle(self, user_gpio, dutycycle):
            self.outputs[user_gpio] = dutycycle / 255

        async def stop(self):
            pass

        async def i2c_open(self, bus, address):
            return bus, address

        async def i2c_close(self, handle):
            return

        async def i2c_write_byte_data(self, handle, register, data):
            """Write byte to i2c register on handle."""
            self.i2c_write.append((handle, register, data))

        async def i2c_read_byte_data(self, handle, register):
            """Write byte to i2c register on handle."""
            return self.i2c_read.pop(0)


class TestRpi(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/rpi/'

    def get_platform(self):
        return False

    def setUp(self):
        self.expected_duration = 1.5
        rpi.apigpio = MockApigpio
        super().setUp()

        self.pi = self.machine.default_platform.pi  # type: MockApigpio.Pi

    def testPlatform(self):
        # check modes
        self.assertEqual({1: MockApigpio.INPUT,
                          2: MockApigpio.OUTPUT,
                          7: MockApigpio.INPUT,
                          23: MockApigpio.OUTPUT,
                          30: MockApigpio.OUTPUT},
                         self.pi.modes)

        # check pull ups
        self.assertEqual({1: MockApigpio.PUD_UP,
                          2: MockApigpio.PUD_OFF,
                          7: MockApigpio.PUD_UP,
                          23: MockApigpio.PUD_OFF,
                          30: MockApigpio.PUD_OFF},
                         self.pi.pull_ups)

        # test switches
        self.assertSwitchState("s_test", True)
        self.assertSwitchState("s_test2", False)

        self.pi.callbacks[1](gpio=1, level=0, tick=123)
        self.machine_run()

        self.assertSwitchState("s_test", False)
        self.assertSwitchState("s_test2", False)

        self.pi.callbacks[1](gpio=1, level=1, tick=123)
        self.machine_run()

        self.assertSwitchState("s_test", True)
        self.assertSwitchState("s_test2", False)

        self.pi.callbacks[1](gpio=1, level=0, tick=127)
        self.pi.callbacks[1](gpio=7, level=1, tick=127)
        self.machine_run()

        self.assertSwitchState("s_test", False)
        self.assertSwitchState("s_test2", True)

        # pulse coil
        self.machine.coils["c_test"].pulse()
        self.machine_run()
        self.assertEqual(1, self.pi.outputs[23])
        self.advance_time_and_run(.022)
        self.assertEqual(1, self.pi.outputs[23])
        self.advance_time_and_run(.002)
        self.assertEqual(0, self.pi.outputs[23])

        # enable coil
        self.machine.coils["c_test_allow_enable"].enable()
        self.machine_run()
        self.assertEqual(1, self.pi.outputs[30])
        self.advance_time_and_run(2)
        self.assertEqual(1, self.pi.outputs[30])

        # disable again
        self.machine.coils["c_test_allow_enable"].disable()
        self.machine_run()
        self.assertEqual(0, self.pi.outputs[30])

        # enable with pwm (10ms pulse and 20% duty)
        self.machine.coils["c_pwm"].enable()
        self.machine_run()
        self.assertEqual(1, self.pi.outputs[2])
        self.advance_time_and_run(.01)
        self.assertEqual(0.2, self.pi.outputs[2])

        # test servo
        self.machine.servos["servo1"].go_to_position(0.2)
        self.machine_run()
        self.assertEqual(1200, self.pi.servos[10])

        self.machine.servos["servo1"].go_to_position(0)
        self.machine_run()
        self.assertEqual(1000, self.pi.servos[10])

        self.machine.servos["servo1"].go_to_position(1.0)
        self.machine_run()
        self.assertEqual(2000, self.pi.servos[10])

        device = self.loop.run_until_complete(self.machine.default_platform.configure_i2c("0-123"))

        device.i2c_write8(43, 1337)
        self.machine_run()
        self.assertEqual(((0, 123), 43, 1337), self.pi.i2c_write[0])
        self.pi.i2c_read.append(1337)
        result = self.loop.run_until_complete(device.i2c_read8(43))
        self.assertEqual(1337, result)



"""Base class for flat config players."""
from typing import List

import abc

from mpf.core.config_player import ConfigPlayer


class FlatConfigPlayer(ConfigPlayer, metaclass=abc.ABCMeta):

    """Flat show players."""

    __slots__ = []  # type: List[str]

    def validate_config_entry(self, settings, name):
        """Validate one entry of this player."""
        config = self._parse_config(settings, name)
        return config

    def get_full_config(self, value):
        """Return full config."""
        for element in value:
            value[element] = super().get_full_config(value[element])
        return value

    @abc.abstractmethod
    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Directly play player."""
        # **kwargs since this is an event callback
        raise NotImplementedError

    @abc.abstractmethod
    def get_express_config(self, value) -> dict:
        """Parse short config version.

        Implements "express" settings for this config_player which is what
        happens when a config is passed as a string instead of a full config
        dict. (This is detected automatically and this method is only called
        when the config is not a dict.)

        For example, the led_player uses the express config to parse a string
        like 'ff0000-f.5s' and translate it into:

        color: 220000
        fade: 500

        Since every config_player is different, this method raises a
        NotImplementedError and most be configured in the child class.

        Args:
        ----
            value: The single line string value from a config file.

        Returns a dictionary (which will then be passed through the config
        validator)
        """
        raise NotImplementedError(self.config_file_section)



"""A dual wound coil which consists of two coils."""
from mpf.core.events import event_handler
from mpf.core.system_wide_device import SystemWideDevice


class DualWoundCoil(SystemWideDevice):

    """An instance of a dual wound coil which consists of two coils."""

    config_section = "dual_wound_coils"  # String of the config section name
    collection = "dual_wound_coils"  # String name of the collection
    class_label = "dual_wound_coil"  # String of the friendly name of the device class

    def __init__(self, machine, name):
        """Initialise a dual wound coil."""
        super().__init__(machine, name)
        # Add this device to the coil section
        self.machine.coils[name] = self

    @event_handler(2)
    def event_enable(self, **kwargs):
        """Event handler for enable event."""
        del kwargs
        self.enable()

    def enable(self):
        """Enable a dual wound coil.

        Pulse main coil and enable hold coil.
        """
        self.config['main_coil'].pulse()
        self.config['hold_coil'].enable()

    @event_handler(1)
    def event_disable(self, **kwargs):
        """Event handler for disable event."""
        del kwargs
        self.disable()

    def disable(self):
        """Disable a driver."""
        self.config['main_coil'].disable()
        self.config['hold_coil'].disable()

    @event_handler(3)
    def event_pulse(self, milliseconds: int = None, power: float = None, **kwargs):
        """Event handler for pulse event."""
        del kwargs
        self.pulse(milliseconds, power)

    def pulse(self, milliseconds: int = None, power: float = None):
        """Pulse this driver.

        Args:
        ----
            milliseconds: The number of milliseconds the driver should be
                enabled for. If no value is provided, the driver will be
                enabled for the value specified in the config dictionary.
            power: A multiplier that will be applied to the default pulse time,
                typically a float between 0.0 and 1.0. (Note this is can only be used
                if milliseconds is also specified.)
        """
        self.config['main_coil'].pulse(milliseconds, power)
        self.config['hold_coil'].pulse(milliseconds, power)



"""MPF plugin for an auditor which records switch events, high scores, shots, etc."""

import logging

from mpf.core.switch_controller import MonitoredSwitchChange
from mpf.devices.shot import Shot

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import
    from typing import Any, Set     # pylint: disable-msg=cyclic-import,unused-import


class Auditor:

    """Writes switch events, regular events, and player variables to an audit log file."""

    __slots__ = ["log", "machine", "switchnames_to_audit", "config", "current_audits", "enabled", "data_manager"]

    def __init__(self, machine: "MachineController") -> None:
        """Initialise auditor.

        Args:
        ----
            machine: A reference to the machine controller object.
        """
        if 'auditor' not in machine.config:
            machine.log.debug('"Auditor:" section not found in machine '
                              'configuration, so the auditor will not be '
                              'used.')
            return

        self.log = logging.getLogger('Auditor')
        self.machine = machine

        self.machine.auditor = self
        self.switchnames_to_audit = set()       # type: Set[str]
        self.config = None                      # type: Any
        self.current_audits = None              # type: Any

        self.enabled = False
        """Attribute that's viewed by other core components to let them know
        they should send auditing events. Set this via the enable() and
        disable() methods.
        """

        self.data_manager = self.machine.create_data_manager('audits')

        self.machine.events.add_handler('init_phase_4', self._initialize)

    def __repr__(self):
        """Return string representation."""
        return '<Auditor>'

    def _load_defaults(self):
        """Load defaults if audits are missing."""
        if not isinstance(self.current_audits, dict):
            self.current_audits = dict()

        # Make sure we have all the sections we need in our audit dict
        if 'switches' not in self.current_audits:
            self.current_audits['switches'] = dict()

        if 'events' not in self.current_audits:
            self.current_audits['events'] = dict()

        if 'player' not in self.current_audits:
            self.current_audits['player'] = dict()

        # Make sure we have all the switches in our audit dict
        for switch in self.machine.switches.values():
            if (switch.name not in self.current_audits['switches'] and
                    'no_audit' not in switch.tags):
                self.current_audits['switches'][switch.name] = 0

        # build the list of switches we should audit
        self.switchnames_to_audit = {x.name for x in self.machine.switches.values()
                                     if 'no_audit' not in x.tags}

        for event in self.config["events"]:
            self.current_audits["events"][event] = 0

        # Make sure we have all the player stuff in our audit dict
        if 'player' in self.config['audit']:
            for item in self.config['player']:
                if item not in self.current_audits['player']:
                    self.current_audits['player'][item] = dict()
                    self.current_audits['player'][item]['top'] = list()
                    self.current_audits['player'][item]['average'] = 0
                    self.current_audits['player'][item]['total'] = 0

    def _set_machine_variables(self):
        """Set machine variables for audits."""
        for category, audits in self.current_audits.items():
            if not isinstance(audits, dict):
                continue
            for name, value in audits.items():
                self.machine.variables.set_machine_var("audits_{}_{}".format(category, name), value)

    def _initialize(self, **kwargs):
        del kwargs
        # Initializes the auditor. We do this separate from __init__() since
        # we need everything else to be setup first.

        self.config = self.machine.config_validator.validate_config('auditor', self.machine.config['auditor'])

        self.current_audits = self.data_manager.get_data()

        self._load_defaults()
        self._set_machine_variables()

        # Register for the events the auditor needs to do its job
        self.machine.events.add_handler('game_starting', self.enable)
        self.machine.events.add_handler('game_ended', self.disable)
        if 'player' in self.config['audit']:
            self.machine.events.add_handler('game_ending', self.audit_player)

        # Enable the shots monitor
        Shot.monitor_enabled = True
        self.machine.register_monitor('shots', self.audit_shot)

        # Add the switches monitor
        self.machine.switch_controller.add_monitor(self.audit_switch)

        for event in self.config['reset_audit_events']:
            self.machine.events.add_handler(event, self._reset)

    def _reset(self, **kwargs):
        """Reset audits."""
        del kwargs
        self.log.info("Resetting audits")
        self.current_audits = {}
        self._load_defaults()
        self._set_machine_variables()
        self._save_audits()

    def audit(self, audit_class, event, **kwargs):
        """Log an auditable event.

        Args:
        ----
            audit_class: A string of the section we want this event to be
                logged to.
            event: A string name of the event we're auditing.
            **kwargs: Not used, but included since some of the audit events
                might include random kwargs.
        """
        del kwargs

        if audit_class not in self.current_audits:
            self.current_audits[audit_class] = dict()

        if event not in self.current_audits[audit_class]:
            self.current_audits[audit_class][event] = 0

        self.current_audits[audit_class][event] += 1
        self.machine.variables.set_machine_var("audits_{}_{}".format(audit_class, event),
                                               self.current_audits[audit_class][event])
        self._save_audits()

    def audit_switch(self, change: MonitoredSwitchChange):
        """Record switch change."""
        if self.enabled and change.state and change.name in self.switchnames_to_audit:
            self.audit('switches', change.name)

    def audit_shot(self, name, profile, state):
        """Record shot hit."""
        del profile
        del state
        self.audit('shots', name)

    def audit_event(self, eventname, **kwargs):
        """Record this event in the audit log.

        Args:
        ----
            eventname: The string name of the event.
            **kwargs: not used, but included since some types of events include
                kwargs.
        """
        del kwargs

        self.current_audits['events'][eventname] += 1
        self._save_audits()

    def audit_player(self, **kwargs):
        """Write player data to the audit log.

        Typically this is only called at the end of a game.

        Args:
        ----
            **kwargs: not used, but included since some types of events include
                kwargs.
        """
        del kwargs
        if not self.machine.game or not self.machine.game.player_list:
            return

        for item in set(self.config['player']):
            for player in self.machine.game.player_list:

                self.current_audits['player'][item]['top'] = (
                    self._merge_into_top_list(
                        player[item],
                        self.current_audits['player'][item]['top'],
                        self.config['num_player_top_records']))

                self.current_audits['player'][item]['average'] = (
                    ((self.current_audits['player'][item]['total'] *
                      self.current_audits['player'][item]['average']) +
                     self.machine.game.player[item]) /
                    (self.current_audits['player'][item]['total'] + 1))

                self.current_audits['player'][item]['total'] += 1
        self._save_audits()

    @classmethod
    def _merge_into_top_list(cls, new_item, current_list, num_items):
        # takes a list of top integers and a new item and merges the new item
        # into the list, then trims it based on the num_items specified
        current_list.append(new_item)
        current_list.sort(reverse=True)
        return current_list[0:num_items]

    def enable(self, **kwargs):
        """Enable the auditor.

        This method lets you enable the auditor so it only records things when
        you want it to. Typically this is called at the beginning of a game.

        Args:
        ----
            **kwargs: No function here. They just exist to allow this method
                to be registered as a handler for events that might contain
                keyword arguments.

        """
        del kwargs
        if self.enabled:
            return  # this will happen if we get a mid game restart

        self.log.debug("Enabling the Auditor")
        self.enabled = True

        # Register for the events we're auditing
        if 'events' in self.config:
            for event in self.config['events']:
                self.machine.events.add_handler(event,
                                                self.audit_event,
                                                eventname=event,
                                                priority=2)
                # Make sure we have an entry in our audit file for this event
                if event not in self.current_audits['events']:
                    self.current_audits['events'][event] = 0

    def _save_audits(self):
        self.data_manager.save_all(data=self.current_audits)

    def disable(self, **kwargs):
        """Disable the auditor."""
        del kwargs
        self.log.debug("Disabling the Auditor")
        self.enabled = False

        # remove switch and event handlers
        self.machine.events.remove_handler(self.audit_event)



"""Contains code for the smart_virtual platform."""
import abc
import logging

from typing import List
from mpf.core.logging import LogMixin
from mpf.core.platform import DriverConfig
from mpf.devices.ball_device.enable_coil_ejector import EnableCoilEjector
from mpf.devices.ball_device.hold_coil_ejector import HoldCoilEjector
from mpf.devices.ball_device.pulse_coil_ejector import PulseCoilEjector

from mpf.platforms.interfaces.driver_platform_interface import PulseSettings, HoldSettings

from mpf.core.delays import DelayManager
from mpf.platforms.virtual import (VirtualHardwarePlatform as VirtualPlatform, VirtualDriver)
from mpf.devices.ball_device.ball_device import BallDevice  # pylint: disable-msg=cyclic-import,unused-import

MYPY = False
if MYPY:   # pragma: no cover
    from typing import Dict     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class BaseSmartVirtualCoilAction(metaclass=abc.ABCMeta):

    """A action for a coil."""

    def __init__(self, actions: List[str], machine: "MachineController") -> None:
        """Initialise switch enable action."""
        self.log = logging.getLogger("SmartVirtual Coil Action")
        self.actions = actions
        self.machine = machine
        self.delay = DelayManager(self.machine)
        self.machine.config['smart_virtual'] = self.machine.config_validator.validate_config(
            "smart_virtual", self.machine.config.get('smart_virtual', {}))

    def enable(self, coil):
        """Enable driver."""
        del coil
        if "enable" in self.actions:
            self._perform_action()
            return True

        return False

    def disable(self, coil):
        """Disable driver."""
        del coil
        if "disable" in self.actions:
            self._perform_action()
            return True

        return False

    def pulse(self, coil, milliseconds):
        """Pulse driver."""
        del coil
        del milliseconds
        if "pulse" in self.actions:
            self._perform_action()
            return True

        return False

    @abc.abstractmethod
    def _perform_action(self):
        """Implement your action here."""


class ResetDropTargetAction(BaseSmartVirtualCoilAction):

    """Set switches inactive when coil is pulsed."""

    def __init__(self, actions, machine, drop_target_bank):
        """Initialise switch enable action."""
        super().__init__(actions, machine)
        self.drop_target_bank = drop_target_bank

    def _hit_switches(self):
        # need to pull the drop targets from the config, not the drop_targets
        # attribute, because if the bank is defined in a mode then the
        # attr won't be populated yet, but if there's a bank in a mode with a
        # single reset coil we assume that coil will reset all the drop targets
        # anytime it's called.
        for target in self.drop_target_bank.config['drop_targets']:
            self.machine.switch_controller.process_switch_obj(target.config['switch'], 0, logical=True)

    def _perform_action(self):
        self.delay.add(ms=50, callback=self._hit_switches)


class SwitchDisableAction(BaseSmartVirtualCoilAction):

    """Disable switches when coil is pulsed."""

    def __init__(self, actions, machine, switches):
        """Initialise switch enable action."""
        super().__init__(actions, machine)
        self.switches = switches

    def _hit_switches(self):
        for switch in self.switches:
            self.log.debug("Disabling switch %s due to coil pulse", switch)
            self.machine.switch_controller.process_switch_obj(switch, 0, logical=True)

    def _perform_action(self):
        self.delay.add(ms=50, callback=self._hit_switches)


class SwitchEnableAction(SwitchDisableAction):

    """Enable switches when coil is pulsed."""

    def _hit_switches(self):
        for switch in self.switches:
            self.log.debug("Enabling switch %s due to coil pulse", switch.name)
            self.machine.switch_controller.process_switch_obj(switch, 1, logical=True)


class ScoreReelAdvanceAction(BaseSmartVirtualCoilAction):

    """Virtual score reel."""

    # pylint: disable-msg=too-many-arguments
    def __init__(self, actions, machine, switch_map, limit_lo, limit_hi, name):
        """Initialise virtual score reel."""
        super().__init__(actions, machine)
        self.switch_map = switch_map
        self.position = limit_lo
        self.limit_lo = limit_lo
        self.limit_hi = limit_hi
        self.name = name

        # set initial position
        self._update_switches()

    def _perform_action(self):
        # increment position and handle overflow
        self.position += 1
        self.position %= self.limit_hi + 1
        if self.position < self.limit_lo:
            self.position = self.limit_lo

        self.machine.log.debug("Virtual score reel for %s at value %s", self.name, self.position)

        self._update_switches()

    def _update_switches(self):
        # disable all switches except the current position
        for position, switch in self.switch_map.items():
            if not switch:
                continue

            state = self.position == position

            if not self.machine.switch_controller.is_state(switch, state):
                self.machine.switch_controller.process_switch_obj(switch, state, logical=True)


class AddBallToTargetAction(BaseSmartVirtualCoilAction):

    """Hit switches when coil is pulsed."""

    # pylint: disable-msg=too-many-arguments
    def __init__(self, actions, machine, platform, device: BallDevice):
        """Initialise add ball to target action."""
        super().__init__(actions, machine)
        self.device = device    # type: BallDevice
        self.ball_switches = device.ball_count_handler.counter.config.get('ball_switches', [])
        self.platform = platform
        self.confirm_eject_switch = device.config['confirm_eject_switch']
        self.target_device = None
        self.result = "success"

    def confirm_eject_via_switch(self, switch):
        """Simulate eject via switch."""
        self.log.debug('Confirming eject via switch: %s', switch.name)
        self.machine.switch_controller.process_switch_obj(switch, 1, logical=True)
        self.delay.add(ms=10, callback=self._release_confirm_switch, switch=switch)

    def _release_confirm_switch(self, switch):
        self.log.debug('Releasing eject confirmation switch: %s', switch.name)
        self.machine.switch_controller.process_switch_obj(switch, 0, logical=True)

    def set_target(self, source, target, mechanical_eject, **kwargs):
        """Set target for action."""
        self.log.debug("Setting eject target. %s -> %s. Mechanical: %s",
                       source.name, target.name, mechanical_eject)
        del kwargs
        driver = None
        if isinstance(source.ejector, (EnableCoilEjector, PulseCoilEjector)) and source.ejector.config['eject_coil']:
            driver = source.ejector.config['eject_coil'].hw_driver
        elif isinstance(source.ejector, HoldCoilEjector) and source.ejector.config['hold_coil']:
            driver = source.ejector.config['hold_coil'].hw_driver
        if driver and driver.action:
            driver.action.target_device = target

        if "delay" in self.actions and self.machine.config['smart_virtual']['simulate_manual_plunger']:
            self.log.debug("Adding %sms delay before ejecting",
                           self.machine.config['smart_virtual']['simulate_manual_plunger_timeout'])

            # simulate mechanical eject
            self.delay.add(ms=self.machine.config['smart_virtual']['simulate_manual_plunger_timeout'],
                           callback=self._perform_action)

    def set_result(self, result):
        """Set result to "success" (default), "return" or "missing"."""
        self.result = result

    def _perform_action(self):
        self.log.debug("Removing ball from device %s", self.device.name)

        if not self.device.balls:
            return

        for switch in self.ball_switches:
            if self.machine.switch_controller.is_active(switch):
                self.machine.switch_controller.process_switch_obj(switch, 0, logical=True)
                self.log.debug("Deactivating: %s", switch.name)
                break

        if self.result == "success":
            if (self.device.ball_count_handler.counter.config.get('entrance_switch_full_timeout') and
                    self.device.machine.switch_controller.is_active(
                        self.device.ball_count_handler.counter.config['entrance_switch'][0])):

                self.machine.switch_controller.process_switch_obj(
                    self.device.ball_count_handler.counter.config['entrance_switch'][0], 0, logical=True)
                self.log.debug("Deactivating: %s",
                               self.device.ball_count_handler.counter.config['entrance_switch'][0].name)

            if self.confirm_eject_switch:
                self.delay.add(ms=50, callback=self.confirm_eject_via_switch,
                               switch=self.confirm_eject_switch)
                self.log.debug("Adding delay for confirm eject switch")

            if self.target_device and not self.target_device.is_playfield():
                self.delay.add(ms=100, callback=self.platform.add_ball_to_device,
                               device=self.target_device)
                self.log.debug("Adding delay for %s to receive ball in 100ms", self.target_device.name)
                self.target_device = None
        elif self.result == "return":
            self.delay.add(ms=100, callback=self.platform.add_ball_to_device,
                           device=self.target_device)
        elif self.result == "missing":
            if self.target_device.is_playfield() and not self.confirm_eject_switch:
                raise AssertionError("Ball cannot go missing to a playfield without confirm_switch.")
            # missing does not do anything
        else:
            raise AssertionError("Invalid result {}".format(self.result))


class SmartVirtualHardwarePlatform(VirtualPlatform):

    """Base class for the smart_virtual hardware platform."""

    def __init__(self, machine):
        """Initialise smart virtual platform."""
        super().__init__(machine)
        self.delay = DelayManager(self.machine)
        self.actions = {}       # type: Dict[BallDevice, AddBallToTargetAction]

    def __repr__(self):
        """Return string representation."""
        return '<Platform.SmartVirtual>'

    def _setup_log(self):
        self.log = logging.getLogger("Smart Virtual Platform")
        self.debug_log("Configuring smart_virtual hardware interface.")

    async def start(self):
        """Initialise platform when all devices are ready."""
        self._initialise_ball_devices()
        self._initialise_drop_targets()
        self._initialise_drop_target_banks()
        self._initialise_score_reels()

    def _initialise_score_reels(self):
        for device in self.machine.score_reels.values():
            if device.config['coil_inc']:
                device.config['coil_inc'].hw_driver.action = ScoreReelAdvanceAction(
                    ["pulse"], self.machine,
                    {
                        0: device.config['switch_0'],
                        1: device.config['switch_1'],
                        2: device.config['switch_2'],
                        3: device.config['switch_3'],
                        4: device.config['switch_4'],
                        5: device.config['switch_5'],
                        6: device.config['switch_6'],
                        7: device.config['switch_7'],
                        8: device.config['switch_8'],
                        9: device.config['switch_9'],
                        10: device.config['switch_10'],
                        11: device.config['switch_11'],
                        12: device.config['switch_12']
                    },
                    device.config['limit_lo'],
                    device.config['limit_hi'],
                    device.name
                )

    def _initialise_drop_targets(self):
        for device in self.machine.drop_targets.values():
            if device.config['reset_coil']:
                device.config['reset_coil'].hw_driver.action = SwitchDisableAction(
                    ["pulse"], self.machine, [device.config['switch']])
            if device.config['knockdown_coil']:
                device.config['knockdown_coil'].hw_driver.action = SwitchEnableAction(
                    ["pulse"], self.machine, [device.config['switch']])

    def _initialise_drop_target_banks(self):
        for device in self.machine.drop_target_banks.values():
            if device.config['reset_coil']:
                device.config['reset_coil'].hw_driver.action = ResetDropTargetAction(
                    ["pulse"], self.machine, device)

            for coil in device.config['reset_coils']:
                coil.hw_driver.action = ResetDropTargetAction(
                    ["pulse"], self.machine, device)

    def _initialise_ball_devices(self):
        for device in self.machine.ball_devices.values():
            if device.is_playfield():
                continue

            action = None
            if isinstance(device.ejector, (EnableCoilEjector, PulseCoilEjector)) and \
                    device.ejector.config['eject_coil']:
                action = device.ejector.config['eject_coil'].hw_driver.action = AddBallToTargetAction(
                    ["pulse", "enable"], self.machine, self, device)

            elif isinstance(device.ejector, HoldCoilEjector) and device.ejector.config['hold_coil']:
                action = device.ejector.config['hold_coil'].hw_driver.action = AddBallToTargetAction(
                    ["disable"], self.machine, self, device)
            elif device.config['mechanical_eject']:
                action = AddBallToTargetAction(["delay"], self.machine, self, device)

            if action:
                # we assume that the device always reaches its target. diverters are ignored
                self.machine.events.add_handler('balldevice_{}_ejecting_ball'.format(device.name),
                                                action.set_target)
                self.actions[device] = action

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict):
        """Configure driver."""
        del platform_settings
        # generate number if None
        if number is None:
            number = str(self._next_driver)
            self._next_driver += 1

        driver = SmartVirtualDriver(config, number)

        return driver

    def add_ball_to_device(self, device: BallDevice):
        """Add ball to device."""
        self.debug_log("Adding ball to %s", device.name)
        if LogMixin.unit_test and device.balls >= device.capacity:
            raise AssertionError("KABOOM! We just added a ball to {} which has a capacity "
                                 "of {} but already had {} ball(s)".format(
                                     device.name, device.capacity, device.balls))

        if device.ball_count_handler.counter.config.get('entrance_switch'):
            entrance_switch = device.ball_count_handler.counter.config['entrance_switch'][0]
            # if there's an entrance_switch_full_timeout, that means the ball
            # will sit on this switch if the device is full, otherwise, it
            # will pass over it, hitting during the process
            if device.ball_count_handler.counter.config.get('entrance_switch_full_timeout'):
                if device.balls == device.capacity - 1:

                    if LogMixin.unit_test and self.machine.switch_controller.is_active(entrance_switch):
                        raise AssertionError(
                            "KABOOM! We just added a ball to {} which already "
                            "had an active entrance switch {}".format(
                                device.name, entrance_switch))

                    self.debug_log('Enabling switch %s due to ball being added to %s',
                                   entrance_switch.name, device.name)

                    self.machine.switch_controller.process_switch_obj(entrance_switch, 1, True)
                    return

            self.debug_log('Hitting switch %s due to ball being added to %s', entrance_switch.name, device.name)
            self.machine.switch_controller.process_switch_obj(entrance_switch, 1, True)
            self.machine.switch_controller.process_switch_obj(entrance_switch, 0, True)

        if device.ball_count_handler.counter.config.get('ball_switches'):
            found_switch = False
            for switch in device.ball_count_handler.counter.config['ball_switches']:
                if self.machine.switch_controller.is_inactive(switch):
                    self.debug_log('Enabling switch %s due to ball being added to %s', switch.name, device.name)
                    self.machine.switch_controller.process_switch_obj(
                        switch, 1, logical=True)
                    found_switch = True
                    break

            if LogMixin.unit_test and not found_switch:
                raise AssertionError("KABOOM! We just added a ball to {} which "
                                     "was already full.".format(device.name))


class SmartVirtualDriver(VirtualDriver):

    """Smart virtual driver."""

    def __init__(self, config, number):
        """Initialise smart virtual driver."""
        super().__init__(config, number)
        self.action = None

    def __repr__(self):
        """Return string representation."""
        return "SmartVirtualDriver.{}".format(self.number)

    def disable(self):
        """Disable driver."""
        super().disable()
        if self.action:
            self.action.disable(self)

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable driver."""
        super().enable(pulse_settings, hold_settings)
        if self.action:
            self.action.enable(self)

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse driver."""
        super().pulse(pulse_settings)
        if self.action:
            self.action.pulse(self, pulse_settings.duration)



from mpf.tests.MpfGameTestCase import MpfGameTestCase


class TestBallDeviceEventEjector(MpfGameTestCase):

    def get_config_file(self):
        return 'test_ball_device_event_ejector.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def testEject(self):
        self.hit_switch_and_run("s_ball_switch1", 1)
        self.hit_switch_and_run("s_ball_switch2", 1)

        self.mock_event("trough_eject")
        self.machine.playfield.add_ball(1)
        self.advance_time_and_run()
        self.assertEventCalled("trough_eject")



"""BCP pickle client."""
import asyncio
import pickle
import struct

from mpf.core.bcp.bcp_client import BaseBcpClient


class BcpPickleClient(BaseBcpClient):

    """BCP client using pickle."""

    def __init__(self, machine, name, bcp):
        """Initialise BCP pickle client."""
        super().__init__(machine, name, bcp)
        self._receiver = None
        self._sender = None

    async def read_message(self):
        """Read the next message."""
        message_length = struct.unpack("!I", (await self._receiver.readexactly(4)))[0]
        message_raw = await self._receiver.readexactly(message_length)

        return pickle.load(message_raw)

    async def connect(self, config):
        """Actively connect to server."""
        config = self.machine.config_validator.validate_config(
            'bcp:connections', config, 'bcp:connections')

        self.info_log("Connecting BCP to '%s' at %s:%s...",
                      self.name, config['host'], config['port'])

        while True:
            connector = self.machine.clock.open_connection(config['host'], config['port'])
            try:
                self._receiver, self._sender = await connector
            except (ConnectionRefusedError, OSError):
                if config.get('required'):
                    await asyncio.sleep(.1)
                    continue

                self.info_log("No BCP connection made to '%s' %s:%s",
                              self.name, config['host'], config['port'])
                return False
            break

        self.info_log("Connected BCP to '%s' %s:%s", self.name, config['host'], config['port'])
        return True

    def accept_connection(self, receiver, sender):
        """Create client for incoming connection."""
        self._receiver = receiver
        self._sender = sender

    def send(self, bcp_command, kwargs):
        """Send message."""
        message_raw = pickle.dump(bcp_command, kwargs)
        complete_message = struct.pack("!I", len(message_raw)) + message_raw

        if hasattr(self._sender.transport, "is_closing") and self._sender.transport.is_closing():
            self.warning_log("Failed to write to bcp since transport is closing. Transport %s", self._sender.transport)
            return
        self._sender.write(complete_message)

    def stop(self):
        """Stop client."""
        self.debug_log("Stopping socket client")
        self._sender.close()



from mpf.tests.MpfGameTestCase import MpfGameTestCase
from mpf.tests.MpfTestCase import test_config


class TestAttract(MpfGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/game/'

    def get_platform(self):
        return 'smart_virtual'

    def testAttractLoadingAndUnloading(self):
        self.assertNumBallsKnown(0)

        self.assertModeRunning("attract")
        self.assertModeNotRunning("game")

        # ball search is not running because we are missing enough balls
        self.assertFalse(self.machine.playfield.ball_search.enabled)
        self.assertFalse(self.machine.playfield.ball_search.started)

        # cannot start game
        self.assertGameIsNotRunning()
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertGameIsNotRunning()
        self.assertModeRunning("attract")
        self.assertModeNotRunning("game")
        # this should start ball search
        self.assertTrue(self.machine.playfield.ball_search.enabled)
        self.assertTrue(self.machine.playfield.ball_search.started)

        self.fill_troughs()
        self.advance_time_and_run()

        # start a three ball one player game
        self.start_game()
        for i in range(3):
            self.advance_time_and_run(1)
            self.assertModeNotRunning("attract")
            self.assertModeRunning("game")
            self.advance_time_and_run(9)
            self.drain_all_balls()

        self.advance_time_and_run(1)

        self.assertModeRunning("attract")
        self.assertModeNotRunning("game")

        self.assertFalse(self.machine.playfield.ball_search.enabled)
        self.assertFalse(self.machine.playfield.ball_search.started)

    @test_config("config_with_balls.yaml")
    def testAttractLoadingWithOneBallInTrough(self):
        self.assertNumBallsKnown(1)

        self.assertModeRunning("attract")
        self.assertModeNotRunning("game")

        # ball search is not running because we are missing enough balls
        self.assertTrue(self.machine.playfield.ball_search.enabled)
        self.assertTrue(self.machine.playfield.ball_search.started)

        # cannot start game
        self.assertGameIsNotRunning()
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertGameIsNotRunning()
        self.assertModeRunning("attract")
        self.assertModeNotRunning("game")

        self.fill_troughs()
        self.advance_time_and_run()

        # start a three ball one player game
        self.start_game()
        for i in range(3):
            self.advance_time_and_run(1)
            self.assertModeNotRunning("attract")
            self.assertModeRunning("game")
            self.advance_time_and_run(9)
            self.drain_all_balls()

        self.advance_time_and_run(1)

        self.assertModeRunning("attract")
        self.assertModeNotRunning("game")



from mpf.tests.MpfTestCase import MpfTestCase, test_config_directory


class TestCustomCode(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/custom_code/'

    def test_scoring(self):
        self.mock_event("test_response")
        self.post_event("test_event")
        self.machine_run()

        self.assertEqual(1, self._events['test_response'])

    @test_config_directory('tests/machine_files/scriptlet/')
    def test_legacy_scriptlet(self):
        self.mock_event("test_response")
        self.post_event("test_event")
        self.machine_run()

        self.assertEqual(1, self._events['test_response'])



import os
from unittest import TestCase

import mpf
from mpf.core.config_processor import ConfigProcessor

from mpf.core.config_loader import YamlMultifileConfigLoader


class TestConfigLoader(TestCase):

    def test_yaml_multifile_config_loader(self):

        machine_path = os.path.abspath(os.path.join(mpf.core.__path__[0], os.pardir,
                                                    "tests/machine_files/config_loader/"))

        config_file = ["config.yaml"]

        config_loader = YamlMultifileConfigLoader(machine_path, config_file, False, False)
        config = config_loader.load_mpf_config()

        self.assertTrue(config.get_config_spec())
        self.assertTrue(config.get_machine_config())
        self.assertTrue(config.get_mode_config("attract"))
        self.assertTrue(config.get_mode_config("game"))
        self.assertTrue(config.get_mode_config("mode1"))
        self.assertTrue(config.get_mode_config("mode2"))
        self.assertTrue(config.get_show_config("mode1_show"))
        self.assertTrue(config.get_show_config("show1"))
        self.assertCountEqual(
            ['flash', 'on', 'off', 'led_color', 'flash_color', 'test_show', 'show1', 'game_show', 'mode1_show'],
            config.get_shows())



"""Test event player."""
from mpf.tests.MpfTestCase import MpfTestCase
from mpf.tests.MpfGameTestCase import MpfGameTestCase

# How many test iterations do we run to convince of randomness?
RANDOM_RUNS = 20


class TestRandomEventPlayer(MpfTestCase):

    def get_config_file(self):
        return 'test_random_event_player.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/event_players/'

    def test_machine_tests(self):
        tester = TestRandomEventPlayerBase(self, "machine")
        tester.run_all_tests()


class TestRandomEventPlayerGame(MpfGameTestCase):

    def get_config_file(self):
        return 'test_random_event_player.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/event_players/'

    def get_platform(self):
        return 'smart_virtual'

    def test_player_tests(self):
        self.fill_troughs()
        self.start_game()
        self.advance_time_and_run(4)
        self.machine.events.post("start_mode2")
        self.advance_time_and_run(4)
        self.assertTrue(self.machine.modes["mode2"].active)
        self.assertTrue(self.machine.mode_controller.is_active('mode2'))

        tester = TestRandomEventPlayerBase(self, "player")
        tester.run_all_tests()


# Below are the common setup and test run methods shared by the machine and player test cases
class TestRandomEventPlayerBase():
    def __init__(self, runner, scope, events=None):
        if events is None:
            events = ["event1", "event2", "event3", "event4"]
        self.runner = runner
        self.scope = scope
        self.events = events
        for event in events:
            runner.machine.events.add_handler(event=event, handler=self._handle_event, event_name=event)

    def run_all_tests(self):
        for test in [self._test_force_different, self._test_force_all, self._test_disable_random]:
            test()

    def _reset(self):
        self.eventHistory = []
        self.uniqueEvents = set()
        self.prevEvent = None
        self.lastEvent = None

    def _handle_event(self, **kwargs):
        self.prevEvent = self.lastEvent
        self.lastEvent = kwargs['event_name']
        self.eventHistory.append(self.lastEvent)
        self.uniqueEvents.add(self.lastEvent)

    def _test_force_different(self):
        self._reset()
        for x in range(0, RANDOM_RUNS):
            self.runner.post_event("test_{}_force_different".format(self.scope))
            self.runner.advance_time_and_run()
            self.runner.assertNotEqual(self.prevEvent, self.lastEvent)

    def _test_force_all(self):
        self._reset()
        for x in range(0, RANDOM_RUNS):
            # If we have done all the events, reset the unique list
            if len(self.events) == len(self.uniqueEvents):
                self.uniqueEvents = set()
                self.eventHistory = []
            self.runner.post_event("test_{}_force_all".format(self.scope))
            self.runner.advance_time_and_run()
            self.runner.assertNotEqual(self.prevEvent, self.lastEvent)
            # If we force all, the set should always get an entry
            self.runner.assertEqual(len(self.uniqueEvents), len(self.eventHistory))

    def _test_disable_random(self):
        self._reset()
        # Set initial values for our lookback
        self.prevEvent = self.events[:1]
        for x in range(0, RANDOM_RUNS):
            expectedIdx = x % 4
            self.runner.post_event("test_{}_disable_random".format(self.scope))
            self.runner.advance_time_and_run()
            self.runner.assertNotEqual(self.prevEvent, self.lastEvent)
            self.runner.assertEqual(self.lastEvent, self.events[expectedIdx])
    
    def _test_conditional_random(self):
        self._reset()
        
        # Only one event is true
        results = list()
        for x in range(RANDOM_RUNS):
            self.runner.post_event("test_{}_conditional_random".format(self.scope))
            self.runner.advance_time_and_run()
            self.assertEqual(RANDOM_RUNS, results.count('event1'))
        
        # Conditional event kwarg is true
        results = list()
        for x in range(RANDOM_RUNS):
            self.runner.post_event("test_{}_conditional_random".format(self.scope), event_arg="foo")
            self.runner.advance_time_and_run()
            self.assertAlmostEqual(RANDOM_RUNS / 2, results.count('event1'), delta=3)
            self.assertAlmostEqual(RANDOM_RUNS / 2, results.count('event3'), delta=3)

        # Machine setting event is true
        self.runner.machine.settings.set_setting_value("foo","bar")
        results = list()
        for x in range(RANDOM_RUNS):
            self.runner.post_event("test_{}_conditional_random".format(self.scope))
            self.runner.advance_time_and_run()
            self.assertAlmostEqual(RANDOM_RUNS / 2, results.count('event1'), delta=3)
            self.assertAlmostEqual(RANDOM_RUNS / 2, results.count('event4'), delta=3)


from mpf.tests.MpfGameTestCase import MpfGameTestCase


class TestPlayerVars(MpfGameTestCase):

    def get_config_file(self):
        return 'player_vars.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/player_vars/'

    def test_initial_values(self):
        self.fill_troughs()
        self.start_two_player_game()

        for x in range(2):

            self.assertEqual(self.machine.game.player_list[x].some_var, 4)
            self.assertEqual(type(self.machine.game.player_list[x].some_var), int)

            self.assertEqual(self.machine.game.player_list[x].some_float, 4.0)
            self.assertEqual(type(self.machine.game.player_list[x].some_float), float)

            self.assertEqual(self.machine.game.player_list[x].some_string, '4')
            self.assertEqual(type(self.machine.game.player_list[x].some_string), str)

            self.assertEqual(self.machine.game.player_list[x].some_other_string, 'hello')
            self.assertEqual(type(self.machine.game.player_list[x].some_other_string), str)

        self.machine.game.player.test = 7
        self.assertEqual(7, self.machine.game.player.test)
        self.assertEqual(7, self.machine.game.player.vars["test"])

        self.assertEqual(4, self.machine.variables.get_machine_var("test1"))
        self.assertEqual('5', self.machine.variables.get_machine_var("test2"))



"""Loads MPF configs."""
from typing import List

import logging
import os
import pickle
import sys

from pathlib import PurePath

from mpf.core.config_processor import ConfigProcessor


class MpfConfig:

    """Contains a MPF config."""

    __slots__ = ["_config_spec", "_machine_config", "_mode_config", "_show_config", "_machine_path", "_mpf_path"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, config_spec, machine_config, modes, shows, machine_path, mpf_path):
        """Initialize config."""
        self._config_spec = config_spec
        self._machine_config = machine_config
        self._mode_config = modes
        self._show_config = shows
        self._machine_path = machine_path
        self._mpf_path = mpf_path

    def get_mpf_path(self):
        """Return mpf path."""
        return self._mpf_path

    def get_machine_path(self):
        """Return machine path."""
        return self._machine_path

    def get_config_spec(self):
        """Return config spec."""
        return self._config_spec

    def get_machine_config(self):
        """Return machine wide config."""
        return self._machine_config

    def get_mode_config(self, mode_name):
        """Return config for a mode."""
        try:
            return self._mode_config[mode_name]
        except KeyError:
            raise AssertionError("No config found for mode '{mode_name}'. MPF expects the config at "
                                 "'modes/{mode_name}/config/{mode_name}.yaml' inside your machine "
                                 "folder.".format(mode_name=mode_name))

    def get_modes(self):
        """Return a list of mode names."""
        return self._mode_config.keys()

    def get_show_config(self, show_name):
        """Return a show."""
        try:
            return self._show_config[show_name]
        except KeyError:
            raise AssertionError("No config found for show '{}'.".format(show_name))

    def get_shows(self):
        """Return a list of all shows names."""
        return self._show_config.keys()


class MpfMcConfig:

    """Contains a MPF-MC config."""

    __slots__ = ["_config_spec", "_machine_config", "_mode_config", "_machine_path"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, config_spec, machine_config, modes, machine_path):
        """Initialize config."""
        self._config_spec = config_spec
        self._machine_config = machine_config
        self._mode_config = modes
        self._machine_path = machine_path

    def get_machine_path(self):
        """Return machine path."""
        return self._machine_path

    def get_config_spec(self):
        """Return config spec."""
        return self._config_spec

    def get_machine_config(self):
        """Return machine wide config."""
        return self._machine_config

    def get_mode_config(self, mode_name):
        """Return config for a mode."""
        try:
            return self._mode_config[mode_name]
        except KeyError:
            raise AssertionError("No config found for mode '{mode_name}'. MPF expects the config at "
                                 "'modes/{mode_name}/config/{mode_name}.yaml' inside your machine "
                                 "folder.".format(mode_name=mode_name))

    def get_modes(self):
        """Return a list of mode names."""
        return self._mode_config.keys()


class ConfigLoader:

    """Generic loader for MPF and MC configs."""

    __slots__ = []  # type: List[str]

    def load_mpf_config(self) -> MpfConfig:
        """Load and return a MPF config."""

    def load_mc_config(self) -> MpfMcConfig:
        """Load and return a MC config."""


class YamlMultifileConfigLoader(ConfigLoader):

    """Loads MPF configs from machine folder with config and modes."""

    __slots__ = ["configfile", "machine_path", "config_processor", "log", "mpf_path", "mc_path"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, machine_path, configfile, load_cache, store_cache):
        """Initialize yaml multifile config loader."""
        self.configfile = configfile
        self.machine_path = machine_path
        self.config_processor = ConfigProcessor(load_cache, store_cache)
        self.log = logging.getLogger("YamlMultifileConfigLoader")
        try:
            # pylint: disable-msg=import-outside-toplevel
            import mpf.core
            self.mpf_path = os.path.abspath(os.path.join(mpf.core.__path__[0], os.pardir))
        except ImportError:
            self.mpf_path = None
        try:
            # pylint: disable-msg=import-outside-toplevel
            import mpfmc.core
            self.mc_path = os.path.abspath(os.path.join(mpfmc.core.__path__[0], os.pardir))
        except ImportError:
            self.mc_path = None

    def load_mpf_config(self) -> MpfConfig:
        """Load and return a MPF config."""
        config_spec = self._load_config_spec()
        machine_config = self._load_mpf_machine_config(config_spec)
        config_spec = self._load_additional_config_spec(config_spec, machine_config)
        mode_config = self._load_modes(machine_config['mpf']['paths']['modes'], config_spec, machine_config)
        show_config = self._load_shows(config_spec, machine_config, mode_config)
        return MpfConfig(config_spec, machine_config, mode_config, show_config, self.machine_path, self.mpf_path)

    def load_mc_config(self) -> MpfMcConfig:
        """Load and return a MC config."""
        config_spec = self._load_config_spec()
        machine_config = self._load_mc_machine_config(config_spec)
        mode_config = self._load_modes(machine_config['mpf-mc']['paths']['modes'], config_spec, machine_config,
                                       ignore_unknown_sections=True)
        return MpfMcConfig(config_spec, machine_config, mode_config, self.machine_path)

    def _load_config_spec(self):
        return self.config_processor.load_config_spec()

    def _load_mpf_machine_config(self, config_spec):
        config_files = [os.path.join(self.mpf_path, "mpfconfig.yaml")]

        for num, config_file in enumerate(self.configfile):
            config_files.append(os.path.join(self.machine_path, "config", config_file))

            self.log.info("Machine config file #%s: %s", num + 1, config_file)

        return self.config_processor.load_config_files_with_cache(config_files, "machine", config_spec=config_spec)

    def _load_mc_machine_config(self, config_spec):
        if not self.mc_path:
            raise AssertionError("Could not import MPF-MC.")
        config_files = [os.path.join(self.mc_path, "mcconfig.yaml")]

        for num, config_file in enumerate(self.configfile):
            config_files.append(os.path.join(self.machine_path, "config", config_file))

            self.log.info("Machine config file #%s: %s", num + 1, config_file)

        return self.config_processor.load_config_files_with_cache(config_files, "machine", config_spec=config_spec)

    def _load_additional_config_spec(self, config_spec, machine_config):
        """Load additional config specs from devices."""
        sys.path.insert(0, self.machine_path)
        config_spec = self.config_processor.load_device_config_specs(config_spec, machine_config)
        sys.path.remove(self.machine_path)
        return config_spec

    def _load_modes(self, mode_path, config_spec, machine_config, ignore_unknown_sections=False):
        mode_config = {}
        for mode in machine_config.get("modes", {}):
            mpf_config_path = os.path.join(self.mpf_path, "modes", mode, 'config', mode + '.yaml')
            machine_config_path = os.path.join(self.machine_path, mode_path, mode, 'config', mode + '.yaml')
            mode_config_files = []
            if os.path.isfile(mpf_config_path):
                self.log.debug("Loading mode %s from %s", mode, mpf_config_path)
                mode_config_files.append(mpf_config_path)
            if os.path.isfile(machine_config_path):
                self.log.debug("Loading mode %s from %s", mode, mpf_config_path)
                mode_config_files.append(machine_config_path)

            if not mode_config_files:
                raise AssertionError("No config found for mode '{mode_name}'. MPF expects the config at "
                                     "'modes/{mode_name}/config/{mode_name}.yaml' inside your machine "
                                     "folder.".format(mode_name=mode))

            config = self.config_processor.load_config_files_with_cache(mode_config_files, "mode",
                                                                        config_spec=config_spec,
                                                                        ignore_unknown_sections=ignore_unknown_sections)

            if "mode" not in config:
                config["mode"] = dict()

            mode_config[mode] = config
        return mode_config

    def _load_shows_in_folder(self, folder, show_configs, config_spec):
        if not os.path.isdir(folder):
            return show_configs
        # ignore temporary files
        ignore_prefixes = (".", "~")
        # do not get fooled by windows or mac garbage
        ignore_files = ("desktop.ini", "Thumbs.db")

        for this_path, _, files in os.walk(folder, followlinks=True):
            relative_path = PurePath(this_path).relative_to(folder)
            for show_file_name in [f for f in files if f.endswith(".yaml") and not f.startswith(ignore_prefixes) and
                                   f != ignore_files]:
                show_name = show_file_name[:-5]
                if show_name in show_configs:
                    raise AssertionError("Duplicate show {}".format(show_name))
                show_config = self.config_processor.load_config_files_with_cache(
                    [os.path.join(folder, str(relative_path), show_file_name)], "show", config_spec=config_spec)
                show_configs[show_name] = show_config
        return show_configs

    def _load_shows(self, config_spec, machine_config, mode_config):
        show_configs = {}
        for show_name, show_config in machine_config.get("shows", {}).items():
            show_configs[show_name] = show_config

        show_configs = self._load_shows_in_folder(os.path.join(self.machine_path, "shows"), show_configs, config_spec)

        mode_path = machine_config['mpf']['paths']['modes']
        for mode_name, config in mode_config.items():
            for show_name, show_config in config.get("shows", {}).items():
                if show_name in show_configs:
                    raise AssertionError("Duplicate show {}".format(show_name))
                show_configs[show_name] = show_config

            show_configs = self._load_shows_in_folder(os.path.join(self.mpf_path, "modes", mode_name, 'shows'),
                                                      show_configs, config_spec)
            show_configs = self._load_shows_in_folder(os.path.join(self.machine_path, mode_path, mode_name, 'shows'),
                                                      show_configs, config_spec)

        return show_configs


class ProductionConfigLoader(ConfigLoader):

    """Loads a single pickled production config each for MPF and MPF-MC."""

    __slots__ = ["machine_path"]

    def __init__(self, machine_path):
        """Initialize production config loader."""
        self.machine_path = machine_path

    @staticmethod
    def get_mpf_bundle_path(machine_path):
        """Return the path for the MPF bundle."""
        return os.path.join(machine_path, "mpf_config.bundle")

    @staticmethod
    def get_mpf_mc_bundle_path(machine_path):
        """Return the path for the MPF bundle."""
        return os.path.join(machine_path, "mpf_mc_config.bundle")

    def load_mpf_config(self) -> MpfConfig:
        """Load and return a MPF config."""
        with open(self.get_mpf_bundle_path(self.machine_path), "rb") as f:
            return pickle.load(f)

    def load_mc_config(self) -> MpfMcConfig:
        """Load and return a MC config."""
        with open(self.get_mpf_mc_bundle_path(self.machine_path), "rb") as f:
            return pickle.load(f)



from mpf.platforms.interfaces.segment_display_platform_interface import FlashingType
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase
from mpf.tests.MpfTestCase import test_config


class TestSegmentDisplay(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/segment_display/'

    @test_config("game.yaml")
    def test_game(self):
        """Test segment displays in a game for the documentation."""
        display1 = self.machine.segment_displays["display1"]
        display2 = self.machine.segment_displays["display2"]
        display3 = self.machine.segment_displays["display3"]
        display4 = self.machine.segment_displays["display4"]
        display5 = self.machine.segment_displays["display5"]

        self.assertEqual("", display1.hw_display.text)
        self.assertEqual("", display2.hw_display.text)
        self.assertEqual("", display3.hw_display.text)
        self.assertEqual("", display4.hw_display.text)
        self.assertEqual("", display5.hw_display.text)
        self.start_game()

        self.assertEqual("0", display1.hw_display.text)
        self.assertEqual(FlashingType.FLASH_ALL, display1.hw_display.flashing)
        self.assertEqual("", display2.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display2.hw_display.flashing)
        self.assertEqual("", display3.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display3.hw_display.flashing)
        self.assertEqual("", display4.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display4.hw_display.flashing)
        self.assertEqual("1", display5.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display5.hw_display.flashing)

        self.add_player()
        self.assertEqual("0", display1.hw_display.text)
        self.assertEqual(FlashingType.FLASH_ALL, display1.hw_display.flashing)
        self.assertEqual("0", display2.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display2.hw_display.flashing)
        self.assertEqual("", display3.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display3.hw_display.flashing)
        self.assertEqual("", display4.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display4.hw_display.flashing)
        self.assertEqual("1", display5.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display5.hw_display.flashing)

        self.machine.game.player.score += 100
        self.advance_time_and_run()
        self.assertEqual("100", display1.hw_display.text)

        self.drain_all_balls()
        self.assertEqual("100", display1.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display1.hw_display.flashing)
        self.assertEqual("0", display2.hw_display.text)
        self.assertEqual(FlashingType.FLASH_ALL, display2.hw_display.flashing)
        self.assertEqual("", display3.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display3.hw_display.flashing)
        self.assertEqual("", display4.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display4.hw_display.flashing)
        self.assertEqual("1", display5.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display5.hw_display.flashing)

        self.machine.game.player.score += 23
        self.advance_time_and_run()
        self.assertEqual("100", display1.hw_display.text)
        self.assertEqual("23", display2.hw_display.text)

        self.drain_all_balls()
        self.assertEqual("100", display1.hw_display.text)
        self.assertEqual(FlashingType.FLASH_ALL, display1.hw_display.flashing)
        self.assertEqual("23", display2.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display2.hw_display.flashing)
        self.assertEqual("", display3.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display3.hw_display.flashing)
        self.assertEqual("", display4.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display4.hw_display.flashing)
        self.assertEqual("2", display5.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display5.hw_display.flashing)

        self.drain_all_balls()
        self.assertEqual("100", display1.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display1.hw_display.flashing)
        self.assertEqual("23", display2.hw_display.text)
        self.assertEqual(FlashingType.FLASH_ALL, display2.hw_display.flashing)
        self.assertEqual("", display3.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display3.hw_display.flashing)
        self.assertEqual("", display4.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display4.hw_display.flashing)
        self.assertEqual("2", display5.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display5.hw_display.flashing)

        self.drain_all_balls()
        self.assertEqual("100", display1.hw_display.text)
        self.assertEqual(FlashingType.FLASH_ALL, display1.hw_display.flashing)
        self.assertEqual("23", display2.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display2.hw_display.flashing)
        self.assertEqual("", display3.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display3.hw_display.flashing)
        self.assertEqual("", display4.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display4.hw_display.flashing)
        self.assertEqual("3", display5.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display5.hw_display.flashing)

        self.drain_all_balls()
        self.assertEqual("100", display1.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display1.hw_display.flashing)
        self.assertEqual("23", display2.hw_display.text)
        self.assertEqual(FlashingType.FLASH_ALL, display2.hw_display.flashing)
        self.assertEqual("", display3.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display3.hw_display.flashing)
        self.assertEqual("", display4.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display4.hw_display.flashing)
        self.assertEqual("3", display5.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display5.hw_display.flashing)

        # game ended
        self.drain_all_balls()
        self.assertGameIsNotRunning()
        self.assertEqual("100", display1.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display1.hw_display.flashing)
        self.assertEqual("23", display2.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display2.hw_display.flashing)
        self.assertEqual("", display3.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display3.hw_display.flashing)
        self.assertEqual("", display4.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display4.hw_display.flashing)
        self.assertEqual("", display5.hw_display.text)
        self.assertEqual(FlashingType.NO_FLASH, display5.hw_display.flashing)

    def test_player(self):
        display1 = self.machine.segment_displays["display1"]
        display2 = self.machine.segment_displays["display2"]

        self.post_event("test_event1")
        self.advance_time_and_run()

        self.assertEqual("HELLO1", display1.hw_display.text)
        self.assertEqual("HELLO2", display2.hw_display.text)

        self.post_event("test_event2")
        self.advance_time_and_run()

        self.assertEqual("", display1.hw_display.text)
        self.assertEqual("HELLO2", display2.hw_display.text)

        self.post_event("test_flashing")
        self.assertEqual(FlashingType.FLASH_ALL, display1.hw_display.flashing)

        self.post_event("test_no_flashing")
        self.assertEqual(FlashingType.NO_FLASH, display1.hw_display.flashing)

        self.post_event("test_event3")
        self.advance_time_and_run()

        self.assertEqual("", display1.hw_display.text)
        self.assertEqual("", display2.hw_display.text)

        self.post_event("test_score")
        self.advance_time_and_run()

        self.assertEqual("1: 0", display1.hw_display.text)
        self.assertEqual("2: 0", display2.hw_display.text)

        self.machine.variables.set_machine_var("test", 42)
        self.advance_time_and_run()

        self.assertEqual("1: 0", display1.hw_display.text)
        self.assertEqual("2: 42", display2.hw_display.text)

        self.start_game()
        self.machine.game.player.score += 100
        self.advance_time_and_run()
        self.assertEqual("1: 100", display1.hw_display.text)
        self.assertEqual("2: 42", display2.hw_display.text)

        self.machine.game.player.score += 23
        self.machine.variables.set_machine_var("test", 1337)
        self.advance_time_and_run()
        self.assertEqual("1: 123", display1.hw_display.text)
        self.assertEqual("2: 1337", display2.hw_display.text)

        self.post_event("test_flash")
        self.advance_time_and_run(.1)
        self.assertEqual("TEST", display1.hw_display.text)
        self.assertEqual("2: 1337", display2.hw_display.text)

        self.advance_time_and_run(2)
        self.assertEqual("1: 123", display1.hw_display.text)
        self.assertEqual("2: 1337", display2.hw_display.text)

        self.machine.modes["mode1"].start()
        self.advance_time_and_run(.1)
        self.assertEqual("MODE1", display1.hw_display.text)
        self.assertEqual("MODE1", display2.hw_display.text)

        self.machine.modes["mode1"].stop()
        self.advance_time_and_run(7)
        self.assertEqual("1: 123", display1.hw_display.text)
        self.assertEqual("2: 1337", display2.hw_display.text)

        self.machine.modes["mode1"].start()
        self.advance_time_and_run(5)
        self.assertEqual("MODE1", display1.hw_display.text)
        self.assertEqual("MODE1", display2.hw_display.text)

        self.advance_time_and_run(5)
        self.assertEqual("MODE1", display1.hw_display.text)
        self.assertEqual("2: 1337", display2.hw_display.text)

    def test_scoring(self):
        display1 = self.machine.segment_displays["display1"]
        display2 = self.machine.segment_displays["display2"]

        # default scoring
        self.post_event("test_score_two_player")

        # one player game
        self.start_game()

        # first display shows score. second empty
        self.assertEqual("0", display1.hw_display.text)
        self.assertEqual("0", display2.hw_display.text)

        # player scores
        self.machine.game.player.score += 42
        self.advance_time_and_run(.01)
        self.assertEqual("42", display1.hw_display.text)
        self.assertEqual("0", display2.hw_display.text)

        # add player
        self.add_player()
        self.advance_time_and_run(.01)
        self.assertEqual("42", display1.hw_display.text)
        self.assertEqual("0", display2.hw_display.text)



from mpf.tests.MpfGameTestCase import MpfGameTestCase


class TestBallDeviceSingle(MpfGameTestCase):

    def get_platform(self):
        return "smart_virtual"

    def get_config_file(self):
        return 'test_single_device.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def testEjectAndDrain(self):
        # start game
        self.assertEqual(1, self.machine.ball_controller.num_balls_known)

        self.start_game()
        self.assertPlayerNumber(1)
        self.assertBallNumber(1)

        self.machine.playfield.add_ball()
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, self.machine.playfield.available_balls)

        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["trough"])
        self.advance_time_and_run()
        self.assertEqual(0, self.machine.playfield.balls)
        self.assertEqual(1, self.machine.playfield.available_balls)

        self.assertPlayerNumber(1)
        self.assertBallNumber(2)
        self.advance_time_and_run(10.5)
        self.assertEqual(1, self.machine.playfield.available_balls)
        self.assertEqual(1, self.machine.playfield.balls)



"""Handles incoming balls."""
import asyncio
from functools import partial
from mpf.devices.switch import Switch

from mpf.core.utility_functions import Util
from mpf.devices.ball_device.ball_device_state_handler import BallDeviceStateHandler


class IncomingBall:

    """One incoming ball."""

    def __init__(self, source, target):
        """Initialise incoming ball."""
        self._timeout_future = asyncio.Future()
        self._confirm_future = asyncio.Future()
        self._can_skip_future = asyncio.Future()
        self._source = source
        self._target = target
        self._external_confirm_future = None
        self._state = "left_device"
        # 1. ejecting (for space calculation) - not implemented yet
        # 2. left_device (can be confirmed)
        # 3. arrived
        # 4. lost (timeouted)

    def set_can_skip(self):
        """Tell other devices that this ball may have skipped the device."""
        self._can_skip_future.set_result(True)

    def wait_for_can_skip(self):
        """Wait until this future can skip."""
        return asyncio.shield(self._can_skip_future)

    @property
    def can_arrive(self):
        """Return true if ball can arrive."""
        return self._state == "left_device" and \
            (not self._external_confirm_future or self._external_confirm_future.done())

    def add_external_confirm_switch(self, switch: Switch):
        """Add external confirm switch."""
        if self._external_confirm_future:
            raise AssertionError("Can only add external confirm once.")

        self._external_confirm_future = self._source.machine.switch_controller.wait_for_switch(switch)
        self._external_confirm_future.add_done_callback(self._external_confirm)

    def add_external_confirm_event(self, event):
        """Add external confirm event."""
        if self._external_confirm_future:
            raise AssertionError("Can only add external confirm once.")

        self._external_confirm_future = self._source.machine.events.wait_for_event(event)
        self._external_confirm_future.add_done_callback(self._external_confirm)

    def _external_confirm(self, future):
        """Handle exteral confirm done."""
        # do not handle canceled external confirm
        if future.cancelled():
            return
        # cancel current timeout
        self._timeout_future.cancel()
        # set up a timeout for ball missing at target
        timeout = self._source.config['ball_missing_timeouts'][self._target] / 1000
        self._timeout_future = asyncio.ensure_future(asyncio.sleep(timeout))
        # set confirmed for source
        self._confirm_future.set_result(True)

    @property
    def source(self):
        """Return source."""
        return self._source

    @property
    def target(self):
        """Return target."""
        return self._target

    def did_not_arrive(self):
        """Ball did not arrive."""
        if not self._state == "left_device":
            return
        self._state = "lost"

        self._timeout_future.cancel()
        self._target.remove_incoming_ball(self)

    def ball_arrived(self):
        """Ball did arrive."""
        if not self._state == "left_device":
            return
        self._state = "arrived"

        if not self._external_confirm_future:
            self._confirm_future.set_result(True)
        self._target.remove_incoming_ball(self)
        self._timeout_future.cancel()

    def wait_for_confirm(self):
        """Wait for confirm."""
        return asyncio.shield(self._confirm_future)

    def wait_for_timeout(self):
        """Wait for timeout."""
        return asyncio.shield(self._timeout_future)

    @property
    def is_timeouted(self):
        """Return true if timeouted."""
        return self._timeout_future.done() and not self._timeout_future.cancelled()


class IncomingBallsHandler(BallDeviceStateHandler):

    """Handles incoming balls and timeouts."""

    __slots__ = ["_incoming_balls", "_has_no_incoming_balls", "_has_incoming_balls", "_is_timeouting"]

    def __init__(self, ball_device):
        """Initialise incoming balls handler."""
        super().__init__(ball_device)
        # list of incoming balls sorted by their expiring time
        self._incoming_balls = []
        self._has_incoming_balls = asyncio.Event()
        self._has_no_incoming_balls = asyncio.Event()
        self._has_no_incoming_balls.set()
        self._is_timeouting = asyncio.Lock()

    def get_num_incoming_balls(self):
        """Return number of incoming ball."""
        return len(self._incoming_balls)

    def wait_for_no_incoming_balls(self):
        """Wait until there are no incoming balls."""
        return self._has_no_incoming_balls.wait()

    async def _run(self):
        """Handle timeouts."""
        while True:
            if not self._incoming_balls:
                self._has_incoming_balls.clear()
                self._has_no_incoming_balls.set()

            # sleep until we have incoming balls
            await self._has_incoming_balls.wait()

            # the incoming balls may have been removed in the meantime
            if not self._incoming_balls:
                continue

            # wait for timeouts on incoming balls
            futures = [incoming_ball.wait_for_timeout() for incoming_ball in self._incoming_balls]
            await Util.first(futures)

            await self._is_timeouting.acquire()

            # handle timeouts
            timeouts = []
            for incoming_ball in self._incoming_balls:
                if incoming_ball.is_timeouted:
                    timeouts.append(incoming_ball)

            for incoming_ball in timeouts:
                self.ball_device.log.warning("Incoming ball from %s timeouted.", incoming_ball.source)
                self._incoming_balls.remove(incoming_ball)

            for incoming_ball in timeouts:
                await self.ball_device.lost_incoming_ball(source=incoming_ball.source)

            self._is_timeouting.release()

    def start_eject(self):
        """Stop counting because eject counter will do that."""
        return self._is_timeouting.acquire()

    def end_eject(self):
        """Restart counting because eject ended."""
        self._is_timeouting.release()

    def add_incoming_ball(self, incoming_ball: IncomingBall):
        """Add incoming balls."""
        self.debug_log("Adding incoming ball from %s", incoming_ball.source)
        self._incoming_balls.append(incoming_ball)
        self._has_incoming_balls.set()
        self._has_no_incoming_balls.clear()

        if self.ball_device.config['mechanical_eject']:
            incoming_ball.wait_for_can_skip().add_done_callback(
                partial(self._add_incoming_ball_which_may_skip_cb, incoming_ball))

    def _add_incoming_ball_which_may_skip_cb(self, incoming_ball, future):
        if not future.cancelled():
            self.ball_device.outgoing_balls_handler.add_incoming_ball_which_may_skip(incoming_ball)

    def remove_incoming_ball(self, incoming_ball: IncomingBall):
        """Remove incoming ball."""
        self.debug_log("Removing incoming ball from %s", incoming_ball.source)
        self._incoming_balls.remove(incoming_ball)
        if self.ball_device.config['mechanical_eject'] and incoming_ball.wait_for_can_skip().done():
            self.ball_device.outgoing_balls_handler.remove_incoming_ball_which_may_skip(incoming_ball)

    async def ball_arrived(self):
        """Handle one ball which arrived in the device."""
        for incoming_ball in self._incoming_balls:
            if not incoming_ball.can_arrive:
                continue

            # handle incoming ball
            self.info_log("Received ball from %s", incoming_ball.source)

            # confirm eject
            incoming_ball.ball_arrived()

            await self.ball_device.expected_ball_received()
            break
        else:
            # handle unexpected ball
            self.info_log("Received unexpected ball")
            # let the ball device handle this ball
            await self.ball_device.unexpected_ball_received()



from unittest.mock import MagicMock
from mpf.tests.MpfTestCase import MpfTestCase


class TestServo(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/servo/'

    def test_servo_go_to_position(self):
        # full range servo
        servo = self.machine.servos["test_servo"]
        # go to position 1.0 (on of the ends)
        servo.go_to_position(1.0)
        # assert that platform got called
        self.assertEqual(1.0, servo.hw_servo.current_position)
        # go to position 0.0 (other end)
        servo.go_to_position(0.0)
        # assert that platform got called
        self.assertEqual(0.0, servo.hw_servo.current_position)

        # limited range servo (like most ones)
        servo = self.machine.servos["limited_servo"]
        # go to position 1.0 (on of the ends)
        servo.go_to_position(1.0)
        # assert that platform got called
        self.assertEqual(0.8, servo.hw_servo.current_position)
        # go to position 0.0 (other end)
        servo.go_to_position(0.0)
        # assert that platform got called
        self.assertEqual(0.2, servo.hw_servo.current_position)
        # go to position 0.0 (middle)
        servo.go_to_position(0.5)
        # assert that platform got called
        self.assertEqual(0.5, servo.hw_servo.current_position)

    def test_events(self):
        servo = self.machine.servos["test_servo"]

        # post reset event
        self.post_event("test_reset")
        # should go to reset position
        self.assertEqual(0.5, servo.hw_servo.current_position)

        # post another defined event
        self.post_event("test_00")
        self.assertEqual(0.0, servo.hw_servo.current_position)

        # post another defined event
        self.post_event("test_01")
        self.assertEqual(0.1, servo.hw_servo.current_position)

        # post another defined event
        self.post_event("test_10")
        self.assertEqual(1.0, servo.hw_servo.current_position)



"""Test placeholders."""
import unittest
from unittest.mock import MagicMock

from mpf.exceptions.config_file_error import ConfigFileError
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase

from mpf.core.placeholder_manager import PlaceholderManager, BoolTemplate, TextTemplate


class TestPlaceholderManager(unittest.TestCase):

    def test_operations(self):
        mock_machine = MagicMock()
        p = PlaceholderManager(mock_machine)

        # test mod operator
        template = p.build_int_template("a % 7", None)
        self.assertEqual(3, template.evaluate({"a": 10}))

        # test "true" and "false"
        template = p.build_int_template("a == true", None)
        with self.assertRaises(ConfigFileError):
            self.assertFalse(False, template.evaluate({"a": True}))

    def test_conditionals(self):
        mock_machine = MagicMock()
        p = PlaceholderManager(mock_machine)

        # test no conditions or numbers
        d = p.parse_conditional_template("test_string")
        self.assertEqual(d.name, "test_string")
        self.assertIsNone(d.condition)
        self.assertIsNone(d.number)
        # test number with pipe
        d = p.parse_conditional_template("test_string|12")
        self.assertEqual(d.name, "test_string")
        self.assertIsNone(d.condition)
        self.assertEqual(d.number, "12")
        # test number with colon
        d = p.parse_conditional_template("test_string:21")
        self.assertEqual(d.name, "test_string")
        self.assertIsNone(d.condition)
        self.assertEqual(d.number, "21")
        # test conditional event
        d = p.parse_conditional_template("test_string{somecondition<1}")
        self.assertEqual(d.name, "test_string")
        self.assertIsInstance(d.condition, BoolTemplate)
        self.assertIsNone(d.number)
        # test conditional event with pipe
        d = p.parse_conditional_template("test_string{somecondition<1}|32")
        self.assertEqual(d.name, "test_string")
        self.assertIsInstance(d.condition, BoolTemplate)
        self.assertEqual(d.number, "32")
        # test conditional event with colon
        d = p.parse_conditional_template("test_string{somecondition<1}:34")
        self.assertEqual(d.name, "test_string")
        self.assertIsInstance(d.condition, BoolTemplate)
        self.assertEqual(d.number, "34")
        # test default number
        d = p.parse_conditional_template("test_string", default_number=2)
        self.assertEqual(d.name, "test_string")
        self.assertEqual(d.number, 2)
        # test number typing
        d = p.parse_conditional_template("test_string|45", default_number=1.0)
        self.assertEqual(d.number, 45)
        self.assertIsInstance(d.number, float)
        # test number without typing
        d = p.parse_conditional_template("test_string:500ms")
        self.assertEqual(d.number, "500ms")
        # test fallback on unparseable number
        d = p.parse_conditional_template("test_string|foobar", default_number=8)
        self.assertEqual(d.number, 8)

class TestPlaceholderManagerWithMachine(MpfFakeGameTestCase):

    def test_subscribe(self):
        self.start_game()
        template = self.machine.placeholder_manager.build_int_template(
            "machine.a + current_player.b", 0)

        value, subscription = template.evaluate_and_subscribe([])
        self.assertFalse(subscription.done())
        self.assertEqual(0, value)

        self.machine.variables.set_machine_var("c", 3)
        self.advance_time_and_run()
        self.assertFalse(subscription.done())

        self.machine.variables.set_machine_var("a", 3)
        self.advance_time_and_run()
        self.assertTrue(subscription.done())

        value, subscription = template.evaluate_and_subscribe([])
        self.assertFalse(subscription.done())
        self.assertEqual(3, value)

        self.machine.game.player.a = 7
        self.advance_time_and_run()
        self.assertFalse(subscription.done())

        self.machine.game.player.b = 7
        self.advance_time_and_run()
        self.assertTrue(subscription.done())

        value, subscription = template.evaluate_and_subscribe([])
        self.assertFalse(subscription.done())
        self.assertEqual(10, value)

        self.machine.game.player.b = 8
        self.advance_time_and_run()

    def test_player_vars(self):
        self.start_game()
        template_game = self.machine.placeholder_manager.build_int_template(
            "game.max_players", 0)
        game = template_game.evaluate([])
        self.assertEqual(4, game)
        template_game = self.machine.placeholder_manager.build_int_template(
            "game.num_players", 0)
        game = template_game.evaluate([])
        self.assertEqual(1, game)

        template_current = self.machine.placeholder_manager.build_int_template(
            "current_player.score", 0)

        template1 = self.machine.placeholder_manager.build_int_template(
            "players[0].score", 0)
        template2 = self.machine.placeholder_manager.build_int_template(
            "players[1].score", 0)

        value, subscription = template_current.evaluate_and_subscribe([])
        value1, subscription1 = template1.evaluate_and_subscribe([])
        value2, subscription2 = template2.evaluate_and_subscribe([])
        self.assertFalse(subscription.done())
        self.assertEqual(0, value)
        self.assertFalse(subscription1.done())
        self.assertEqual(0, value1)
        self.assertFalse(subscription2.done())
        self.assertEqual(0, value2)

        self.machine.game.player.score += 100
        self.advance_time_and_run(.1)
        self.assertTrue(subscription.done())
        self.assertTrue(subscription1.done())
        value, subscription = template_current.evaluate_and_subscribe([])
        value1, subscription1 = template1.evaluate_and_subscribe([])
        value2, subscription2 = template2.evaluate_and_subscribe([])
        self.assertFalse(subscription.done())
        self.assertEqual(100, value)
        self.assertFalse(subscription1.done())
        self.assertEqual(100, value1)
        self.assertFalse(subscription2.done())
        self.assertEqual(0, value2)

        self.add_player()
        self.advance_time_and_run(.1)
        self.assertTrue(subscription2.done())
        value, subscription = template_current.evaluate_and_subscribe([])
        value1, subscription1 = template1.evaluate_and_subscribe([])
        value2, subscription2 = template2.evaluate_and_subscribe([])
        self.assertFalse(subscription.done())
        self.assertEqual(100, value)
        self.assertFalse(subscription1.done())
        self.assertEqual(100, value1)
        self.assertFalse(subscription2.done())
        self.assertEqual(0, value2)

        template_game = self.machine.placeholder_manager.build_int_template(
            "game.num_players", 0)
        game = template_game.evaluate([])
        self.assertEqual(2, game)

        self.drain_all_balls()
        self.advance_time_and_run(.1)
        value, subscription = template_current.evaluate_and_subscribe([])
        value1, subscription1 = template1.evaluate_and_subscribe([])
        value2, subscription2 = template2.evaluate_and_subscribe([])
        self.assertFalse(subscription.done())
        self.assertEqual(0, value)
        self.assertFalse(subscription1.done())
        self.assertEqual(100, value1)
        self.assertFalse(subscription2.done())
        self.assertEqual(0, value2)

        self.machine.game.player.score += 42
        self.advance_time_and_run(.1)
        self.assertTrue(subscription2.done())
        value, subscription = template_current.evaluate_and_subscribe([])
        value1, subscription1 = template1.evaluate_and_subscribe([])
        value2, subscription2 = template2.evaluate_and_subscribe([])
        self.assertFalse(subscription.done())
        self.assertEqual(42, value)
        self.assertFalse(subscription1.done())
        self.assertEqual(100, value1)
        self.assertFalse(subscription2.done())
        self.assertEqual(42, value2)
        self.advance_time_and_run(.1)
        self.assertFalse(subscription1.done())
        self.assertFalse(subscription2.done())

        self.machine.game.player_list[0].score += 23
        self.advance_time_and_run(.1)
        self.assertTrue(subscription1.done())
        value, subscription = template_current.evaluate_and_subscribe([])
        value1, subscription1 = template1.evaluate_and_subscribe([])
        value2, subscription2 = template2.evaluate_and_subscribe([])
        self.assertFalse(subscription.done())
        self.assertEqual(42, value)
        self.assertFalse(subscription1.done())
        self.assertEqual(123, value1)
        self.assertFalse(subscription2.done())
        self.assertEqual(42, value2)

    def testTextTemplate(self):
        t = TextTemplate(self.machine, "Number: {test:<4d}")
        self.assertEqual("Number: 7   ", t.evaluate({"test": 7}))
        self.assertEqual("Number: 0   ", t.evaluate({"test": None}))




"""Scaffold YAMLs in your machine."""
import cmd
import os

from mpf.file_interfaces.yaml_roundtrip import YamlRoundtrip


class ScaffoldCli(cmd.Cmd):

    """Scaffolding for MPF."""

    intro = 'Welcome to the MPF scaffolding shell. Type help or ? to list commands.\n'
    prompt = "(scaffold) "

    def __init__(self, machine_path, stdin=None, stdout=None):
        """Initialise service cli."""
        super().__init__(stdin=stdin, stdout=stdout)
        self.machine_path = machine_path

    def _create_mode(self, name):
        """Create a mode."""
        # create mode folder
        os.mkdir(os.path.join(self.machine_path, "modes", name))
        # create config folder
        os.mkdir(os.path.join(self.machine_path, "modes", name, "config"))
        # create config file
        with open(os.path.join(self.machine_path, "modes", name, "config", "{}.yaml".format(name)), "w") as f:
            f.write("""#config_version=5

mode:
    start_events: ball_started
    priority: 100
""")

        self.stdout.write("Success: Created mode {}. Next step: Add it to your mode list.\n".format(name))

    @staticmethod
    def _create_show(name):
        del name
        raise AssertionError("Not implemented")

    def do_copy(self, args):
        """Copy light positions from monitor to your config."""
        arguments = args.split(" ")
        if not arguments or len(arguments) != 2 or arguments[0] != "light_positions":
            self.stdout.write("Usage: copy light_positions your_light_config_file\n")
            return

        config_loader = YamlRoundtrip()
        config_name = arguments[1]
        try:
            monitor_config = config_loader.load("monitor/monitor.yaml")
        except Exception as e:  # pylint: disable-msg=broad-except
            self.stdout.write("Error while loading monitor/monitor.yaml: {}.\n".format(e))
            return

        try:
            lights_config = config_loader.load(config_name)
        except Exception as e:  # pylint: disable-msg=broad-except
            self.stdout.write("Error while loading {}: {}.\n".format(config_name, e))
            return

        if "light" not in monitor_config:
            self.stdout.write("Error: Monitor config does not contain a light section.\n")
            return

        if "lights" not in lights_config:
            self.stdout.write("Error: Config does not contain a lights section.\n")
            return

        lights_found = 0
        for light_name, light_config in lights_config['lights'].items():
            if light_name in monitor_config['light']:
                monitor_light_config = monitor_config['light'][light_name]
                lights_found += 1

                light_config['x'] = monitor_light_config['x']
                light_config['y'] = monitor_light_config['y']

        config_loader.save(config_name, lights_config)
        self.stdout.write("Success: Found {} lights.\n".format(lights_found))

    def do_create(self, args):
        """Create something."""
        arguments = args.split(" ")
        if not arguments:
            self.stdout.write("Usage: create mode/show [...]\n")
            return
        what = arguments.pop(0)
        if what == "mode":
            if len(arguments) != 1:
                self.stdout.write("Usage: create mode name\n")
                return
            self._create_mode(arguments[0])
        elif what == "show":
            if len(arguments) != 1:
                self.stdout.write("Usage: create show name\n")
                return
            self._create_show(arguments[0])
        else:
            self.stdout.write("Error: Unknown {}\n".format(what))
            return

    def do_exit(self, args):    # noqa
        """Exit scaffolding mode."""
        del args
        return True

    def do_quit(self, args):    # noqa
        """Exit scaffolding mode."""
        del args
        return True

    def do_EOF(self, args):     # noqa
        """Exit scaffolding mode."""
        del args
        return True


class Command:

    """Run the service cli."""

    def __init__(self, mpf_path, machine_path, args):
        """Run service cli."""
        del mpf_path
        del args

        cli = ScaffoldCli(machine_path)
        try:
            cli.cmdloop()
        except KeyboardInterrupt:
            # catch ctrl+c
            cli.do_exit("")

        # print a newline for a nice exit
        print()



"""Device that implements an extra ball."""
from typing import Optional

from mpf.core.events import event_handler
from mpf.core.machine import MachineController
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.mode import Mode
from mpf.core.mode_device import ModeDevice
from mpf.core.player import Player
from mpf.devices.extra_ball_group import ExtraBallGroup


@DeviceMonitor("enabled")
class ExtraBall(ModeDevice):

    """An extra ball which can be awarded once per player."""

    config_section = 'extra_balls'
    collection = 'extra_balls'
    class_label = 'extra_ball'

    __slots__ = ["player", "group"]

    def __init__(self, machine: MachineController, name: str) -> None:
        """Initialise extra ball."""
        super().__init__(machine, name)
        self.player = None  # type: Optional[Player]
        """The current player"""
        self.group = None   # type: Optional[ExtraBallGroup]
        """The ExtraBallGroup this ExtraBall belongs to, or None."""

    @property
    def enabled(self):
        """Return whether this extra ball group is enabled.

        This takes into consideration the enabled setting plus the max balls
        per game setting.
        """
        return self.is_ok_to_award()

    async def _initialize(self):
        await super()._initialize()
        self.group = self.config['group']

    @event_handler(2)
    def event_light(self, **kwargs):
        """Handle light control event."""
        del kwargs
        self.light()

    def light(self):
        """Light an extra ball for potential collection by the player.

        Lighting an extra ball will immediately increase count against the
        ``max_per_game`` setting, even if the extra ball is a member of a
        group that's disabled or if the player never actually collects the
        extra ball.

        Note that this only really does anything if this extra ball is a
        member of a group.
        """
        if self.is_ok_to_light():
            self.machine.events.post('extra_ball_{}_lit'.format(self.name))
            '''event: extra_ball_(name)_lit
            desc: The extra ball called (name) has just been lit.
            '''

            if self.group:
                self.group.light()
        else:
            self._award_disabled()

    @event_handler(1)
    def event_award(self, **kwargs):
        """Handle award control event."""
        del kwargs
        self.award()

    def award(self):
        """Award extra ball to player (if enabled)."""
        if self.is_ok_to_award():
            self.player['extra_ball_{}_num_awarded'.format(self.name)] += 1
            self.machine.events.post('extra_ball_{}_awarded'.format(self.name))
            '''event: extra_ball_(name)_awarded
            desc: The extra ball called (name) has just been awarded.
            '''

            try:
                self.group.award()
            except AttributeError:
                # If this EB is in a group, the group will handle this stuff
                self.player.extra_balls += 1
                self.machine.events.post('extra_ball_awarded')
                '''event: extra_ball_awarded
                desc: An extra ball has just been awarded.
                '''

        else:  # EB cannot be awarded
            self._award_disabled()

    def is_ok_to_light(self) -> bool:
        """Check whether this extra ball can be lit.

        This method takes into consideration whether this extra ball is
        enabled, and, if this extra ball is a member of a group, whether the
        group is enabled and will allow an additional extra ball to lit.

        Returns True or False.
        """
        if self.is_ok_to_award():
            if self.group:
                if self.group.is_ok_to_light():
                    return True
            else:
                return True

        return False

    def is_ok_to_award(self) -> bool:
        """Check whether this extra ball can be awarded.

        This method takes into consideration whether this extra ball is
        enabled, whether the ``max_per_game`` has been exceeded, and, if this
        extra ball is a member of a group, whether the group is enabled and
        will allow an additional extra ball to be awarded.

        Returns True or False.
        """
        if not self.config['enabled'] or not self.player:
            return False

        if self.group and not self.group.enabled:
            return False

        if self.config['max_per_game'] and (
                self.config['max_per_game'] <=
                self.player['extra_ball_{}_num_awarded'.format(self.name)]):
            return False

        return True

    def _award_disabled(self):
        self.machine.events.post('extra_ball_award_disabled')
        '''event: extra_ball_award_disabled
        desc: The award for an extra ball has just been disabled.
        '''

        self.machine.events.post(
            'extra_ball_{}_award_disabled'.format(self.name))
        '''event: extra_ball_(name)_award_disabled
        desc: The award for the extra ball called (name) has just been disabled.
        '''

        if self.group:
            # still need to send this even if EBs are disabled since we
            # want to post the group disabled event
            self.group.award_disabled()

    def device_loaded_in_mode(self, mode: Mode, player: Player):
        """Load extra ball in mode and initialise player.

        Args:
        ----
            mode: Mode which is loaded
            player: Current player
        """
        del mode
        self.player = player

        if not player.is_player_var(
                'extra_ball_{}_num_awarded'.format(self.name)):
            player['extra_ball_{}_num_awarded'.format(self.name)] = 0

        '''player_var: extra_ball_(name)_awarded

        desc: The number of times this extra ball has been awarded to the
        player in this game. Note that the default max is one (meaning that
        each extra ball can be awarded once per game), so this value will only
        be 0 or 1 unless you change the max setting for this extra ball.'''

    def device_removed_from_mode(self, mode: Mode):
        """Unload extra ball.

        Args:
        ----
            mode: Mode which is unloaded
        """
        del mode
        self.player = None



"""Implements a servo in MPF."""
from mpf.core.delays import DelayManager

from mpf.core.device_monitor import DeviceMonitor
from mpf.core.events import event_handler
from mpf.core.platform import ServoPlatform
from mpf.core.system_wide_device import SystemWideDevice


@DeviceMonitor(_position="position")
class Servo(SystemWideDevice):

    """Represents a servo in a pinball machine.

    Args: Same as the Device parent class.
    """

    config_section = 'servos'
    collection = 'servos'
    class_label = 'servo'

    def __init__(self, machine, name):
        """Initialise servo."""
        self.hw_servo = None
        self.platform = None        # type: ServoPlatform
        self._position = None
        self.speed_limit = None
        self.acceleration_limit = None
        self._ball_search_started = False
        self.delay = DelayManager(machine)
        super().__init__(machine, name)

    async def _initialize(self):
        await super()._initialize()
        self.platform = self.machine.get_platform_sections('servo_controllers', self.config['platform'])
        self.platform.assert_has_feature("servos")

        for position in self.config['positions']:
            self.machine.events.add_handler(self.config['positions'][position],
                                            self._position_event,
                                            position=position)

        if not self.platform.features['allow_empty_numbers'] and self.config['number'] is None:
            self.raise_config_error("Servo must have a number.", 1)

        self.hw_servo = await self.platform.configure_servo(self.config['number'])
        self._position = self.config['reset_position']
        self.speed_limit = self.config['speed_limit']
        self.acceleration_limit = self.config['acceleration_limit']

        if self.config['include_in_ball_search']:
            self.machine.events.add_handler("ball_search_started",
                                            self._ball_search_start)
            self.machine.events.add_handler("ball_search_stopped",
                                            self._ball_search_stop)

        self.set_speed_limit(self.speed_limit)
        self.set_acceleration_limit(self.acceleration_limit)

    @event_handler(1)
    def event_reset(self, **kwargs):
        """Event handler for reset event."""
        del kwargs
        self.reset()

    def reset(self):
        """Go to reset position."""
        self.go_to_position(self.config['reset_position'])

    @event_handler(5)
    def _position_event(self, position, **kwargs):
        del kwargs
        self.go_to_position(position)

    def go_to_position(self, position):
        """Move servo to position."""
        self._position = position
        if self._ball_search_started:
            return
        self._go_to_position(position)

    def _go_to_position(self, position):
        # linearly interpolate between servo limits
        corrected_position = self.config['servo_min'] + position * (
            self.config['servo_max'] - self.config['servo_min'])

        self.debug_log("Moving to position %s (corrected: %s)", position, corrected_position)

        # call platform with calculated position
        self.hw_servo.go_to_position(corrected_position)

    def set_speed_limit(self, speed_limit):
        """Set speed parameter."""
        self.hw_servo.set_speed_limit(speed_limit)

    def set_acceleration_limit(self, acceleration_limit):
        """Set acceleration parameter."""
        self.hw_servo.set_acceleration_limit(acceleration_limit)

    def _ball_search_start(self, **kwargs):
        del kwargs
        # we do not touch self._position during ball search so we can reset to
        # it later
        self._ball_search_started = True
        self._ball_search_go_to_min()

    def _ball_search_go_to_min(self):
        self._go_to_position(self.config['ball_search_min'])
        self.delay.add(name="ball_search", callback=self._ball_search_go_to_max, ms=self.config['ball_search_wait'])

    def _ball_search_go_to_max(self):
        self._go_to_position(self.config['ball_search_max'])
        self.delay.add(name="ball_search", callback=self._ball_search_go_to_min, ms=self.config['ball_search_wait'])

    def _ball_search_stop(self, **kwargs):
        del kwargs
        # stop delay
        self.delay.remove("ball_search")
        self._ball_search_started = False

        # move to last position set
        self._go_to_position(self._position)
